from __future__ import annotations

import datetime
import pickle
import re

import numpy as np
import pandas as pd
import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
import ibis.selectors as s
from ibis import _
from ibis.common.annotations import ValidationError
from ibis.common.deferred import Deferred
from ibis.common.exceptions import ExpressionError, IntegrityError, RelationError
from ibis.expr import api
from ibis.expr.rewrites import simplify
from ibis.expr.tests.test_newrels import join_tables
from ibis.expr.types import Column, Table
from ibis.tests.util import assert_equal, assert_pickle_roundtrip


@pytest.fixture
def set_ops_schema_top():
    return [("key", "string"), ("value", "double")]


@pytest.fixture
def set_ops_schema_bottom():
    return [("key", "string"), ("key2", "string"), ("value", "double")]


@pytest.fixture
def setops_table_foo(set_ops_schema_top):
    return ibis.table(set_ops_schema_top, "foo")


@pytest.fixture
def setops_table_bar(set_ops_schema_top):
    return ibis.table(set_ops_schema_top, "bar")


@pytest.fixture
def setops_table_baz(set_ops_schema_bottom):
    return ibis.table(set_ops_schema_bottom, "baz")


@pytest.fixture
def setops_relation_error_message():
    return "Table schemas must be equal for set operations"


def test_empty_schema():
    table = api.table([], "foo")
    assert not table.schema()


def test_columns(con):
    t = con.table("alltypes")
    result = t.columns
    expected = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k"]
    assert result == expected


def test_view_new_relation(table):
    # For assisting with self-joins and other self-referential operations
    # where we need to be able to treat instances of the same Table as
    # semantically distinct
    #
    # This thing is not exactly a projection, since it has no semantic
    # meaning when it comes to execution
    tview1 = table.view()
    tview2 = table.view()
    tview2_ = tview2.view()

    node1 = tview1.op()
    node2 = tview2.op()
    node2_ = tview2_.op()

    assert isinstance(node1, ops.SelfReference)
    assert isinstance(node2, ops.SelfReference)
    assert node1.parent is node2.parent
    assert node1 != node2
    assert node2_ is node2


def test_getitem_column_select(table):
    for k in table.columns:
        col = table[k]

        # Make sure it's the right type
        assert isinstance(col, Column)


def test_select_using_selector(table):
    expr = table[s.numeric()]
    expected = table.select(
        table.a,
        table.b,
        table.c,
        table.d,
        table.e,
        table.f,
    )
    assert expr.equals(expected)


def test_table_tab_completion():
    table = ibis.table({"a": "int", "b": "int", "for": "int", "with spaces": "int"})
    # Only valid python identifiers in getattr tab completion
    attrs = set(dir(table))
    assert {"a", "b"}.issubset(attrs)
    assert {"for", "with spaces"}.isdisjoint(attrs)
    # All columns in getitem tab completion
    items = set(table._ipython_key_completions_())
    assert items.issuperset(table.columns)


def test_getitem_attribute(table):
    result = table.a
    assert_equal(result, table["a"])

    # Project and add a name that conflicts with a Table built-in
    # attribute
    view = table[[table, table["a"].name("schema")]]
    assert not isinstance(view.schema, Column)


def test_getitem_missing_column(table):
    with pytest.raises(com.IbisTypeError, match="oops"):
        table["oops"]


def test_getattr_missing_column(table):
    with pytest.raises(AttributeError, match="oops"):
        table.oops  # noqa: B018


def test_typo_method_name_recommendation(table):
    with pytest.raises(AttributeError, match="order_by"):
        table.sort("a")

    # Existing columns take precedence over raising an error
    # for a common method typo
    table2 = table.rename(sort="a")
    assert isinstance(table2.sort, Column)


def test_projection(table):
    cols = ["f", "a", "h"]

    proj = table[cols]
    assert isinstance(proj, Table)
    assert isinstance(proj.op(), ops.Project)

    assert proj.schema().names == tuple(cols)
    for c in cols:
        expr = proj[c]
        assert isinstance(expr, type(table[c]))


def test_projection_no_list(table):
    expr = (table.f * 2).name("bar")
    result = table.select(expr)
    expected = table.select([expr])
    assert_equal(result, expected)


def test_projection_with_exprs(table):
    # unnamed expr to test
    mean_diff = (table["a"] - table["c"]).mean()

    col_exprs = [table["b"].log().name("log_b"), mean_diff.name("mean_diff")]

    proj = table[col_exprs + ["g"]]
    schema = proj.schema()
    assert schema.names == ("log_b", "mean_diff", "g")
    assert schema.types == (dt.double, dt.double, dt.string)

    # Test with unnamed expr
    proj = table.select(["g", table["a"] - table["c"]])
    schema = proj.schema()
    assert schema.names == ("g", "Subtract(a, c)")
    assert schema.types == (dt.string, dt.int64)


def test_projection_duplicate_names(table):
    with pytest.raises(com.IntegrityError):
        table.select([table.c, table.c])


def test_projection_sort_keys_errors(table):
    """Forbid using `asc`/`desc` in selections"""
    with pytest.raises(ValidationError):
        table.select([table.c.desc()])

    with pytest.raises(ValidationError):
        table.mutate(new=table.c.asc())


def test_projection_invalid_root(table):
    schema1 = {"foo": "double", "bar": "int32"}

    left = api.table(schema1, name="foo")
    right = api.table(schema1, name="bar")

    exprs = [right["foo"], right["bar"]]
    with pytest.raises(IntegrityError):
        left.select(exprs)


def test_projection_with_star_expr(table):
    new_expr = (table["a"] * 5).name("bigger_a")

    t = table

    # it lives!
    proj = t[t, new_expr]
    repr(proj)

    ex_names = table.schema().names + ("bigger_a",)
    assert proj.schema().names == ex_names

    # cannot pass an invalid table expression
    t2 = t.aggregate([t["a"].sum().name("sum(a)")], by=["g"])
    with pytest.raises(IntegrityError):
        t[[t2]]
    # TODO: there may be some ways this can be invalid


def test_projection_convenient_syntax(table):
    proj = table[table, table["a"].name("foo")]
    proj2 = table[[table, table["a"].name("foo")]]
    assert_equal(proj, proj2)


def test_projection_mutate_analysis_bug(con):
    # GH #549

    t = con.table("airlines")

    filtered = t[t.depdelay.notnull()]
    leg = ibis.literal("-").join([t.origin, t.dest])
    mutated = filtered.mutate(leg=leg)

    # it works!
    mutated["year", "month", "day", "depdelay", "leg"]


def test_projection_self(table):
    result = table[table]
    expected = table.select(table)

    assert_equal(result, expected)


def test_projection_array_expr(table):
    result = table[table.a]
    expected = table[[table.a]]
    assert_equal(result, expected)


@pytest.mark.parametrize("empty", [list(), dict()])
def test_projection_no_expr(table, empty):
    with pytest.raises(com.IbisTypeError, match="must select at least one"):
        table.select(empty)


# FIXME(kszucs): currently bind() flattens the list of expressions, so arbitrary
# nesting is allowed, need to revisit
# def test_projection_invalid_nested_list(table):
#     errmsg = "must be coerceable to expressions"
#     with pytest.raises(com.IbisTypeError, match=errmsg):
#         table.select(["a", ["b"]])
#     with pytest.raises(com.IbisTypeError, match=errmsg):
#         table[["a", ["b"]]]
#     with pytest.raises(com.IbisTypeError, match=errmsg):
#         table["a", ["b"]]


def test_mutate(table):
    expr = table.mutate(
        [
            (table.a + 1).name("x1"),
            table.b.sum().name("x2"),
            (_.a + 2).name("x3"),
            lambda _: (_.a + 3).name("x4"),
            4,
            "five",
        ],
        kw1=(table.a + 6),
        kw2=table.b.sum(),
        kw3=(_.a + 7),
        kw4=lambda _: (_.a + 8),
        kw5=9,
        kw6="ten",
    )
    expected = table[
        table,
        (table.a + 1).name("x1"),
        table.b.sum().name("x2"),
        (table.a + 2).name("x3"),
        (table.a + 3).name("x4"),
        ibis.literal(4).name("4"),
        ibis.literal("five").name("'five'"),
        (table.a + 6).name("kw1"),
        table.b.sum().name("kw2"),
        (table.a + 7).name("kw3"),
        (table.a + 8).name("kw4"),
        ibis.literal(9).name("kw5"),
        ibis.literal("ten").name("kw6"),
    ]
    assert_equal(expr, expected)


def test_mutate_alter_existing_columns(table):
    new_f = table.f * 2
    foo = table.d * 2
    expr = table.mutate(f=new_f, foo=foo)

    expected = table[
        "a",
        "b",
        "c",
        "d",
        "e",
        new_f.name("f"),
        "g",
        "h",
        "i",
        "j",
        "k",
        foo.name("foo"),
    ]

    assert_equal(expr, expected)


def test_replace_column():
    tb = api.table([("a", "int32"), ("b", "double"), ("c", "string")])

    expr = tb.b.cast("int32")
    tb2 = tb.mutate(b=expr)
    expected = tb[tb.a, expr.name("b"), tb.c]

    assert_equal(tb2, expected)


def test_filter_no_list(table):
    pred = table.a > 5

    result = table.filter(pred)
    expected = table[pred]
    assert_equal(result, expected)


def test_add_predicate(table):
    pred = table["a"] > 5
    result = table[pred]
    assert isinstance(result.op(), ops.Filter)


def test_invalid_predicate(table, schema):
    # a lookalike
    table2 = api.table(schema, name="bar")
    predicate = table2.a > 5
    with pytest.raises(IntegrityError):
        table.filter(predicate)


def test_add_predicate_coalesce(table):
    # Successive predicates get combined into one rather than nesting. This
    # is mainly to enhance readability since we could handle this during
    # expression evaluation anyway.
    pred1 = table["a"] > 5
    pred2 = table["b"] > 0

    result = simplify(table[pred1][pred2].op()).to_expr()
    expected = table.filter([pred1, pred2])
    assert_equal(result, expected)

    # 59, if we are not careful, we can obtain broken refs
    subset = table[pred1]
    result = simplify(subset.filter([subset["b"] > 0]).op()).to_expr()
    assert_equal(result, expected)


def test_repr_same_but_distinct_objects(con):
    t = con.table("test1")
    t_copy = con.table("test1")
    table2 = t[t_copy["f"] > 0]

    result = repr(table2)
    assert result.count("DatabaseTable") == 1


def test_filter_fusion_distinct_table_objects(con):
    t = con.table("test1")
    tt = con.table("test1")

    expr = t[t.f > 0][t.c > 0]
    expr2 = t[t.f > 0][tt.c > 0]
    expr3 = t[tt.f > 0][tt.c > 0]
    expr4 = t[tt.f > 0][t.c > 0]

    assert_equal(expr, expr2)
    assert repr(expr) == repr(expr2)
    assert_equal(expr, expr3)
    assert_equal(expr, expr4)


def test_relabel():
    table = api.table({"x": "int32", "y": "string", "z": "double"})

    # Using a mapping
    with pytest.warns(FutureWarning, match="Table.rename"):
        res = table.relabel({"x": "x_1", "y": "y_1"}).schema()
    sol = sch.schema({"x_1": "int32", "y_1": "string", "z": "double"})
    assert_equal(res, sol)

    # Using a function
    with pytest.warns(FutureWarning, match="Table.rename"):
        res = table.relabel(lambda x: None if x == "z" else f"{x}_1").schema()
    assert_equal(res, sol)

    # Using a format string
    with pytest.warns(FutureWarning, match="Table.rename"):
        res = table.relabel("_{name}_")
        sol = table.relabel({"x": "_x_", "y": "_y_", "z": "_z_"})
    assert_equal(res, sol)

    # Mapping with unknown columns errors
    with pytest.raises(com.IbisTypeError, match="'missing' is not found in table"):
        with pytest.warns(FutureWarning, match="Table.rename"):
            table.relabel({"missing": "oops"})


def test_rename():
    table = api.table({"x": "int32", "y": "string", "z": "double"})
    sol = sch.schema({"x_1": "int32", "y_1": "string", "z": "double"})

    # Using kwargs
    res = table.rename(x_1="x", y_1="y").schema()
    assert_equal(res, sol)

    # Using a mapping
    res = table.rename({"x_1": "x", "y_1": "y"}).schema()
    assert_equal(res, sol)

    # Using a mix
    res = table.rename({"x_1": "x"}, y_1="y").schema()
    assert_equal(res, sol)


def test_rename_function():
    table = api.table({"x": "int32", "y": "string", "z": "double"})

    res = table.rename(lambda x: None if x == "z" else f"{x}_1").schema()
    sol = sch.schema({"x_1": "int32", "y_1": "string", "z": "double"})
    assert_equal(res, sol)

    # Explicit rename takes precedence
    res = table.rename(lambda x: f"{x}_1", z_2="z").schema()
    sol = sch.schema({"x_1": "int32", "y_1": "string", "z_2": "double"})
    assert_equal(res, sol)


def test_rename_format_string():
    t = ibis.table({"x": "int", "y": "int", "z": "int"})

    res = t.rename("_{name}_")
    sol = t.rename({"_x_": "x", "_y_": "y", "_z_": "z"})
    assert_equal(res, sol)

    with pytest.raises(ValueError, match="Format strings must"):
        t.rename("no format string parameter")

    with pytest.raises(ValueError, match="Format strings must"):
        t.rename("{unknown} format string parameter")


def test_rename_snake_case():
    cases = [
        ("cola", "cola"),
        ("col_b", "ColB"),
        ("col_c", "colC"),
        ("col_d", "col-d"),
        ("col_e", "col_e"),
        ("column_f", " Column F "),
        ("column_g_with_hyphens", "Column G-with-hyphens"),
        ("col_h_notcamelcase", "Col H notCamelCase"),
    ]
    t = ibis.table({c: "int" for _, c in cases})
    res = t.rename("snake_case")
    sol = t.rename(dict(cases))
    assert_equal(res, sol)


def test_rename_all_caps():
    cases = [
        ("COLA", "cola"),
        ("COL_B", "ColB"),
        ("COL_C", "colC"),
        ("COL_D", "col-d"),
        ("COL_E", "col_e"),
        ("COLUMN_F", " Column F "),
        ("COLUMN_G_WITH_HYPHENS", "Column G-with-hyphens"),
        ("COL_H_NOTCAMELCASE", "Col H notCamelCase"),
    ]
    t = ibis.table({c: "int" for _, c in cases})
    res = t.rename("ALL_CAPS")
    sol = t.rename(dict(cases))
    assert_equal(res, sol)


def test_limit(table):
    limited = table.limit(10, offset=5)
    assert limited.op().n == 10
    assert limited.op().offset == 5


def test_order_by(table):
    result = table.order_by(["f"]).op()

    sort_key = result.keys[0]

    assert_equal(sort_key.expr, table.f.op())
    assert sort_key.ascending

    # non-list input. per #150
    result2 = table.order_by("f").op()
    assert_equal(result, result2)

    key2 = result2.keys[0]
    assert key2.descending is False


def test_order_by_desc_deferred_sort_key(table):
    result = table.group_by("g").size().order_by(ibis._[1].desc())

    tmp = table.group_by("g").size()
    expected = tmp.order_by(ibis.desc(tmp[1]))

    assert_equal(result, expected)


def test_order_by_asc_deferred_sort_key(table):
    result = table.group_by("g").size().order_by(ibis._[1])

    tmp = table.group_by("g").size()
    expected = tmp.order_by(tmp[1])
    expected2 = tmp.order_by(ibis.asc(tmp[1]))

    assert_equal(result, expected)
    assert_equal(result, expected2)


@pytest.mark.parametrize(
    ("key", "expected"),
    [
        param(ibis.NA, ibis.NA.op(), id="na"),
        param(ibis.random(), ibis.random().op(), id="random"),
        param(1.0, ibis.literal(1.0).op(), id="float"),
        param(ibis.literal("a"), ibis.literal("a").op(), id="string"),
        param(ibis.literal([1, 2, 3]), ibis.literal([1, 2, 3]).op(), id="array"),
    ],
)
def test_order_by_scalar(table, key, expected):
    result = table.order_by(key)
    assert result.op().keys == (ops.SortKey(expected),)


@pytest.mark.parametrize(
    ("key", "exc_type"),
    [
        ("bogus", com.IbisTypeError),
        # (("bogus", False), com.IbisTypeError),
        (ibis.desc("bogus"), com.IbisTypeError),
        (1000, IndexError),
        # ((1000, False), IndexError),
        (_.bogus, AttributeError),
        (_.bogus.desc(), AttributeError),
    ],
)
@pytest.mark.parametrize(
    "expr_func",
    [
        param(lambda t: t, id="table"),
        param(lambda t: t.select("a", "b"), id="selection"),
        param(lambda t: t.group_by("a").agg(new=_.b.sum()), id="aggregation"),
    ],
)
def test_order_by_nonexistent_column_errors(table, expr_func, key, exc_type):
    # `order_by` is implemented on a few different operations, we check them
    # all in turn here.
    expr = expr_func(table)
    with pytest.raises(exc_type):
        expr.order_by(key)


def test_slice(table):
    expr1 = table[:5]
    expr2 = table[:5:1]
    expr3 = table[5:]
    assert_equal(expr1, table.limit(5))
    assert_equal(expr1, expr2)
    assert_equal(expr3, table.limit(None, offset=5))

    expr1 = table[2:7]
    expr2 = table[2:7:1]
    expr3 = table[2::1]
    assert_equal(expr1, table.limit(5, offset=2))
    assert_equal(expr1, expr2)
    assert_equal(expr3, table.limit(None, offset=2))


@pytest.mark.parametrize("step", [-1, 0, 2])
def test_invalid_slice(table, step):
    with pytest.raises(ValueError):
        table[:5:step]


def test_table_count(table):
    result = table.count()
    assert isinstance(result, ir.IntegerScalar)
    assert isinstance(result.op(), ops.CountStar)


def test_len_raises_expression_error(table):
    with pytest.raises(com.ExpressionError):
        len(table)


def test_sum_expr_basics(table, int_col):
    # Impala gives bigint for all integer types
    result = table[int_col].sum()
    assert isinstance(result, ir.IntegerScalar)
    assert isinstance(result.op(), ops.Sum)


def test_sum_expr_basics_floats(table, float_col):
    # Impala gives double for all floating point types
    result = table[float_col].sum()
    assert isinstance(result, ir.FloatingScalar)
    assert isinstance(result.op(), ops.Sum)


def test_mean_expr_basics(table, numeric_col):
    result = table[numeric_col].mean()
    assert isinstance(result, ir.FloatingScalar)
    assert isinstance(result.op(), ops.Mean)


def test_aggregate_no_keys(table):
    metrics = [
        table["a"].sum().name("sum(a)"),
        table["c"].mean().name("mean(c)"),
    ]

    # A Table, which in SQL at least will yield a table with a single
    # row
    result = table.aggregate(metrics)
    assert isinstance(result, Table)


def test_aggregate_keys_basic(table):
    metrics = [
        table["a"].sum().name("sum(a)"),
        table["c"].mean().name("mean(c)"),
    ]

    # A Table, which in SQL at least will yield a table with a single
    # row
    result = table.aggregate(metrics, by=["g"])
    assert isinstance(result, Table)

    # it works!
    repr(result)


def test_aggregate_having_implicit_metric(table):
    metric = table.f.sum().name("total")
    by = "g"
    having = table.c.sum() > 10

    implicit_having_metric = table.aggregate(metric, by=by, having=having)
    expected_aggregate = ops.Aggregate(
        parent=table,
        groups={"g": table.g},
        metrics={"total": table.f.sum(), table.c.sum().get_name(): table.c.sum()},
    )
    expected_filter = ops.Filter(
        parent=expected_aggregate,
        predicates=[
            ops.Greater(ops.Field(expected_aggregate, table.c.sum().get_name()), 10)
        ],
    )
    expected_project = ops.Project(
        parent=expected_filter,
        values={
            "g": ops.Field(expected_filter, "g"),
            "total": ops.Field(expected_filter, "total"),
        },
    )
    assert implicit_having_metric.op() == expected_project


def test_agg_having_explicit_metric(table):
    metric = table.f.sum().name("total")
    by = "g"
    having = table.c.sum() > 10

    explicit_having_metric = table.aggregate(
        [metric, table.c.sum().name("sum")], by=by, having=having
    )
    expected_aggregate = ops.Aggregate(
        parent=table,
        groups={"g": table.g},
        metrics={"total": table.f.sum(), "sum": table.c.sum()},
    )
    expected_filter = ops.Filter(
        parent=expected_aggregate,
        predicates=[ops.Greater(ops.Field(expected_aggregate, "sum"), 10)],
    )
    assert explicit_having_metric.op() == expected_filter


def test_aggregate_keywords(table):
    t = table

    expr = t.aggregate(foo=t.f.sum(), bar=lambda x: x.f.mean(), by="g")
    expr2 = t.group_by("g").aggregate(foo=t.f.sum(), bar=lambda x: x.f.mean())
    expected = t.aggregate([t.f.sum().name("foo"), t.f.mean().name("bar")], by="g")

    assert_equal(expr, expected)
    assert_equal(expr2, expected)


def test_select_on_literals(table):
    # literal ints and strings are column indices, everything else is a value
    expr1 = table.select(col1=True, col2=1, col3="a")
    expr2 = table.select(col1=ibis.literal(True), col2=table.b, col3=table.a)
    assert expr1.equals(expr2)


def test_filter_on_literal_boolean(table):
    expr1 = table.filter(True)
    expr2 = table.filter(ibis.literal(True))
    assert expr1.equals(expr2)


def test_filter_on_literal_string_is_column(table):
    expr1 = table.filter("h")
    expr2 = table.filter(table.h)
    assert expr1.equals(expr2)


def test_filter_on_literal_then_aggregate(table):
    # Mostly just a smoketest, this used to error on construction
    expr = table.filter(ibis.literal(True)).agg(lambda t: t.a.sum().name("total"))
    assert expr.columns == ["total"]


def test_group_by_having_api(table):
    # #154, add a HAVING post-predicate in a composable way
    metric = table.f.sum().name("foo")
    postp = table.d.mean() > 1
    expr = table.group_by("g").having(postp).aggregate(metric)

    agg = ops.Aggregate(
        parent=table,
        groups={"g": table.g},
        metrics={"foo": table.f.sum(), "Mean(d)": table.d.mean()},
    ).to_expr()
    filt = ops.Filter(
        parent=agg,
        predicates=[agg["Mean(d)"] > 1],
    ).to_expr()
    proj = ops.Project(
        parent=filt,
        values={"g": filt.g, "foo": filt.foo},
    )
    assert expr.op() == proj


def test_group_by_kwargs(table):
    t = table
    expr = t.group_by(["f", t.h], z="g", z2=t.d).aggregate(t.d.mean().name("foo"))
    expected = t.group_by(["f", t.h, t.g.name("z"), t.d.name("z2")]).aggregate(
        t.d.mean().name("foo")
    )
    assert_equal(expr, expected)


def test_compound_aggregate_expr(table):
    # See ibis #24
    compound_expr = (table["a"].sum() / table["a"].mean()).name("foo")

    # Validates internally
    table.aggregate([compound_expr])


def test_groupby_convenience(table):
    metrics = [table.f.sum().name("total")]

    expr = table.group_by("g").aggregate(metrics)
    expected = table.aggregate(metrics, by=["g"])
    assert_equal(expr, expected)

    group_expr = table.g.cast("double").name("g")
    expr = table.group_by(group_expr).aggregate(metrics)
    expected = table.aggregate(metrics, by=[group_expr])
    assert_equal(expr, expected)


@pytest.mark.parametrize("group", [[], (), None])
def test_group_by_nothing(table, group):
    with pytest.raises(com.IbisInputError):
        table.group_by(group)


def test_group_by_count_size(table):
    # #148, convenience for interactive use, and so forth
    result1 = table.group_by("g").size()
    result2 = table.group_by("g").count()

    expected = table.group_by("g").aggregate(table.count())

    assert_equal(result1, expected)
    assert_equal(result2, expected)


def test_group_by_column_select_api(table):
    grouped = table.group_by("g")

    result = grouped.f.sum()
    expected = grouped.aggregate(table.f.sum().name("sum(f)"))
    assert_equal(result, expected)

    supported_functions = ["sum", "mean", "count", "size", "max", "min"]

    # make sure they all work
    for fn in supported_functions:
        getattr(grouped.f, fn)()


def test_value_counts_convenience(table):
    # #152
    result = table.g.value_counts()
    expected = table.select("g").group_by("g").aggregate(g_count=lambda t: t.count())

    assert_equal(result, expected)


def test_isin_value_counts(table):
    # #157, this code path was untested before
    bool_clause = table.g.notin(["1", "4", "7"])
    # it works!
    bool_clause.name("notin").value_counts()


def test_value_counts_unnamed_expr(con):
    nation = con.table("tpch_nation")

    expr = nation.n_name.lower().value_counts()
    expected = nation.n_name.lower().name("Lowercase(n_name)").value_counts()
    assert_equal(expr, expected)


def test_aggregate_unnamed_expr(con):
    nation = con.table("tpch_nation")
    expr = nation.n_name.lower().left(1)

    agg = nation.group_by(expr).aggregate(nation.count().name("metric"))
    schema = agg.schema()
    assert schema.names == ("Substring(Lowercase(n_name), 0, 1)", "metric")
    assert schema.types == (dt.string, dt.int64)


def test_join_no_predicate_list(con):
    region = con.table("tpch_region")
    nation = con.table("tpch_nation")

    pred = region.r_regionkey == nation.n_regionkey
    joined = region.inner_join(nation, pred)

    with join_tables(region, nation) as (r1, r2):
        expected = ops.JoinChain(
            first=r1,
            rest=[ops.JoinLink("inner", r2, [r1.r_regionkey == r2.n_regionkey])],
            values={
                "r_regionkey": r1.r_regionkey,
                "r_name": r1.r_name,
                "r_comment": r1.r_comment,
                "n_nationkey": r2.n_nationkey,
                "n_name": r2.n_name,
                "n_regionkey": r2.n_regionkey,
                "n_comment": r2.n_comment,
            },
        )
        assert joined.op() == expected


def test_join_deferred(con):
    region = con.table("tpch_region")
    nation = con.table("tpch_nation")

    res = region.join(nation, _.r_regionkey == nation.n_regionkey)

    with join_tables(region, nation) as (r1, r2):
        expected = ops.JoinChain(
            first=r1,
            rest=[ops.JoinLink("inner", r2, [r1.r_regionkey == r2.n_regionkey])],
            values={
                "r_regionkey": r1.r_regionkey,
                "r_name": r1.r_name,
                "r_comment": r1.r_comment,
                "n_nationkey": r2.n_nationkey,
                "n_name": r2.n_name,
                "n_regionkey": r2.n_regionkey,
                "n_comment": r2.n_comment,
            },
        )
        assert res.op() == expected


def test_join_invalid_predicate(con):
    region = con.table("tpch_region")
    nation = con.table("tpch_nation")

    with pytest.raises(com.InputTypeError):
        region.inner_join(nation, object())


def test_asof_join():
    left = ibis.table([("time", "int32"), ("value", "double")])
    right = ibis.table([("time", "int32"), ("value2", "double")])
    joined = api.asof_join(left, right, "time")

    assert joined.columns == [
        "time",
        "value",
        "time_right",
        "value2",
    ]
    pred = joined.op().rest[0].predicates[0]
    assert pred.left.name == pred.right.name == "time"


# TODO(kszucs): ensure the correctness of the pd.merge_asof(by=...) argument emulation
def test_asof_join_with_by():
    left = ibis.table([("time", "int32"), ("key", "int32"), ("value", "double")])
    right = ibis.table([("time", "int32"), ("key", "int32"), ("value2", "double")])

    join_without_by = api.asof_join(left, right, "time")
    with join_tables(left, right) as (r1, r2):
        r2 = join_without_by.op().rest[0].table.to_expr()
        expected = ops.JoinChain(
            first=r1,
            rest=[ops.JoinLink("asof", r2, [r1.time >= r2.time])],
            values={
                "time": r1.time,
                "key": r1.key,
                "value": r1.value,
                "time_right": r2.time,
                "key_right": r2.key,
                "value2": r2.value2,
            },
        )
        assert join_without_by.op() == expected

    join_with_by = api.asof_join(left, right, "time", by="key")
    with join_tables(left, right) as (r1, r2):
        expected = ops.JoinChain(
            first=r1,
            rest=[
                ops.JoinLink("asof", r2, [r1.time >= r2.time, r1.key == r2.key]),
            ],
            values={
                "time": r1.time,
                "key": r1.key,
                "value": r1.value,
                "time_right": r2.time,
                "key_right": r2.key,
                "value2": r2.value2,
            },
        )
        assert join_with_by.op() == expected


@pytest.mark.parametrize(
    ("ibis_interval", "timedelta_interval"),
    [
        [ibis.interval(days=2), pd.Timedelta("2 days")],
        [ibis.interval(days=2), datetime.timedelta(days=2)],
        [ibis.interval(hours=5), pd.Timedelta("5 hours")],
        [ibis.interval(hours=5), datetime.timedelta(hours=5)],
        [ibis.interval(minutes=7), pd.Timedelta("7 minutes")],
        [ibis.interval(minutes=7), datetime.timedelta(minutes=7)],
        [ibis.interval(seconds=9), pd.Timedelta("9 seconds")],
        [ibis.interval(seconds=9), datetime.timedelta(seconds=9)],
        [ibis.interval(milliseconds=11), pd.Timedelta("11 milliseconds")],
        [ibis.interval(milliseconds=11), datetime.timedelta(milliseconds=11)],
        [ibis.interval(microseconds=15), pd.Timedelta("15 microseconds")],
        [ibis.interval(microseconds=15), datetime.timedelta(microseconds=15)],
        [ibis.interval(nanoseconds=17), pd.Timedelta("17 nanoseconds")],
    ],
)
def test_asof_join_with_tolerance(ibis_interval, timedelta_interval):
    left = ibis.table([("time", "timestamp"), ("key", "int32"), ("value", "double")])
    right = ibis.table([("time", "timestamp"), ("key", "int32"), ("value2", "double")])

    for interval in [ibis_interval, timedelta_interval]:
        joined = api.asof_join(left, right, "time", tolerance=interval)

        asof = left.asof_join(right, "time")
        filt = asof.filter(
            [
                asof.time <= asof.time_right + interval,
                asof.time >= asof.time_right - interval,
            ]
        )
        join = left.left_join(filt, [left.time == filt.time])
        expected = join.select(
            left,
            time_right=filt.time_right,
            key_right=filt.key_right,
            value2=filt.value2,
        )

        assert joined.equals(expected)


def test_equijoin_schema_merge():
    table1 = ibis.table([("key1", "string"), ("value1", "double")])
    table2 = ibis.table([("key2", "string"), ("stuff", "int32")])

    pred = table1["key1"] == table2["key2"]
    join_types = ["inner_join", "left_join", "outer_join"]

    ex_schema = ibis.schema(
        names=["key1", "value1", "key2", "stuff"],
        types=["string", "double", "string", "int32"],
    )

    for fname in join_types:
        f = getattr(table1, fname)
        joined = f(table2, [pred])
        assert_equal(joined.schema(), ex_schema)


def test_join_combo_with_projection(table):
    # Test a case where there is column name overlap, but the projection
    # passed makes it a non-issue. Highly relevant with self-joins
    #
    # For example, where left/right have some field names in common:
    # SELECT left.*, right.a, right.b
    # FROM left join right on left.key = right.key
    t = table
    t2 = t.mutate(foo=t.f * 2, bar=t.f * 4)

    # this works
    joined = t.left_join(t2, [t["g"] == t2["g"]])
    proj = joined.select([t, t2["foo"], t2["bar"]])
    repr(proj)


def test_join_getitem_projection(con):
    region = con.table("tpch_region")
    nation = con.table("tpch_nation")

    pred = region.r_regionkey == nation.n_regionkey
    joined = region.inner_join(nation, pred)

    result = joined[nation]
    expected = joined.select(nation)
    assert_equal(result, expected)


def test_self_join(table):
    # Self-joins are problematic with this design because column
    # expressions may reference either the left or right  For example:
    #
    # SELECT left.key, sum(left.value - right.value) as total_deltas
    # FROM table left
    #  INNER JOIN table right
    #    ON left.current_period = right.previous_period + 1
    # GROUP BY 1
    #
    # One way around the self-join issue is to force the user to add
    # prefixes to the joined fields, then project using those. Not that
    # satisfying, though.
    left = table
    right = table.view()
    metric = (left["a"] - right["b"]).mean().name("metric")

    joined = left.inner_join(right, [right["g"] == left["g"]])

    # Project out left table schema
    proj = joined[[left]]
    assert_equal(proj.schema(), left.schema())

    # Try aggregating on top of joined
    aggregated = joined.aggregate([metric], by=[left["g"]])
    ex_schema = api.Schema({"g": "string", "metric": "double"})
    assert_equal(aggregated.schema(), ex_schema)


def test_self_join_no_view_convenience(table):
    # #165, self joins ought to be possible when the user specifies the
    # column names to join on rather than referentially-valid expressions

    result = table.join(table, [("g", "g")])
    expected_cols = list(table.columns)
    # TODO(kszucs): the inner join convenience to don't duplicate the
    # equivalent columns from the right table is not implemented yet
    expected_cols.extend(f"{c}_right" for c in table.columns if c != "g")
    assert result.columns == expected_cols


def test_join_reference_bug(con):
    # GH#403
    orders = con.table("tpch_orders")
    customer = con.table("tpch_customer")
    lineitem = con.table("tpch_lineitem")

    items = orders.join(lineitem, orders.o_orderkey == lineitem.l_orderkey)[
        lineitem, orders.o_custkey, orders.o_orderpriority
    ].join(customer, [("o_custkey", "c_custkey")])
    items["o_orderpriority"].value_counts()


def test_join_project_after(table):
    # e.g.
    #
    # SELECT L.foo, L.bar, R.baz, R.qux
    # FROM table1 L
    #   INNER JOIN table2 R
    #     ON L.key = R.key
    #
    # or
    #
    # SELECT L.*, R.baz
    # ...
    #
    # The default for a join is selecting all fields if possible
    table1 = ibis.table([("key1", "string"), ("value1", "double")])
    table2 = ibis.table([("key2", "string"), ("stuff", "int32")])

    pred = table1["key1"] == table2["key2"]

    joined = table1.left_join(table2, [pred])
    projected = joined.select([table1, table2["stuff"]])
    assert projected.schema().names == ("key1", "value1", "stuff")

    projected = joined.select([table2, table1["key1"]])
    assert projected.schema().names == ("key2", "stuff", "key1")


def test_semi_join_schema(table):
    # A left semi join discards the schema of the right table
    table1 = ibis.table([("key1", "string"), ("value1", "double")])
    table2 = ibis.table([("key2", "string"), ("stuff", "double")])

    pred = table1["key1"] == table2["key2"]
    semi_joined = table1.semi_join(table2, [pred])

    result_schema = semi_joined.schema()
    assert_equal(result_schema, table1.schema())


def test_cross_join(table):
    metrics = [
        table["a"].sum().name("sum_a"),
        table["b"].mean().name("mean_b"),
    ]
    scalar_aggs = table.aggregate(metrics)

    joined = table.cross_join(scalar_aggs)
    agg_schema = api.Schema({"sum_a": "int64", "mean_b": "double"})
    ex_schema = table.schema() | agg_schema
    assert_equal(joined.schema(), ex_schema)


def test_cross_join_multiple(table):
    a = table["a", "b", "c"]
    b = table["d", "e"]
    c = table["f", "h"]

    joined = ibis.cross_join(a, b, c)
    with join_tables(a, b, c) as (r1, r2, r3):
        expected = ops.JoinChain(
            first=r1,
            rest=[
                ops.JoinLink("cross", r2, []),
                ops.JoinLink("cross", r3, []),
            ],
            values={
                "a": r1.a,
                "b": r1.b,
                "c": r1.c,
                "d": r2.d,
                "e": r2.e,
                "f": r3.f,
                "h": r3.h,
            },
        )
        assert joined.op() == expected
        # TODO(kszucs): it must be simplified first using an appropriate rewrite rule
        assert not joined.equals(a.cross_join(b.cross_join(c)))


def test_filter_join():
    table1 = ibis.table({"key1": "string", "key2": "string", "value1": "double"})
    table2 = ibis.table({"key3": "string", "value2": "double"})

    # It works!
    joined = table1.inner_join(table2, [table1["key1"] == table2["key3"]])
    filtered = joined.filter([table1.value1 > 0])
    repr(filtered)


def test_inner_join_overlapping_column_names():
    t1 = ibis.table([("foo", "string"), ("bar", "string"), ("value1", "double")])
    t2 = ibis.table([("foo", "string"), ("bar", "string"), ("value2", "double")])

    joined = t1.join(t2, "foo")
    expected = t1.join(t2, t1.foo == t2.foo)
    assert_equal(joined, expected)
    assert joined.columns == ["foo", "bar", "value1", "bar_right", "value2"]

    joined = t1.join(t2, ["foo", "bar"])
    expected = t1.join(t2, [t1.foo == t2.foo, t1.bar == t2.bar])
    assert_equal(joined, expected)
    assert joined.columns == ["foo", "bar", "value1", "value2"]

    # Equality predicates don't have same name, need to rename
    joined = t1.join(t2, t1.foo == t2.bar)
    assert joined.columns == [
        "foo",
        "bar",
        "value1",
        "foo_right",
        "bar_right",
        "value2",
    ]

    # Not all predicates are equality, still need to rename
    joined = t1.join(t2, ["foo", t1.value1 < t2.value2])
    assert joined.columns == [
        "foo",
        "bar",
        "value1",
        "foo_right",
        "bar_right",
        "value2",
    ]


@pytest.mark.parametrize(
    "key_maker",
    [
        lambda t1, t2: t1.foo_id == t2.foo_id,
        lambda t1, t2: [("foo_id", "foo_id")],
        lambda t1, t2: [(t1.foo_id, t2.foo_id)],
        lambda t1, t2: [(_.foo_id, _.foo_id)],
        lambda t1, t2: [(t1.foo_id, _.foo_id)],
        lambda t1, t2: [(2, 0)],  # foo_id is 2nd in t1, 0th in t2
        lambda t1, t2: [(lambda t: t.foo_id, lambda t: t.foo_id)],
    ],
)
def test_join_key_alternatives(con, key_maker):
    t1 = con.table("star1")
    t2 = con.table("star2")
    key = key_maker(t1, t2)

    joined = t1.inner_join(t2, key)
    with join_tables(t1, t2) as (r1, r2):
        expected = ops.JoinChain(
            first=r1,
            rest=[
                ops.JoinLink("inner", r2, [r1.foo_id == r2.foo_id]),
            ],
            values={
                "c": r1.c,
                "f": r1.f,
                "foo_id": r1.foo_id,
                "bar_id": r1.bar_id,
                "value1": r2.value1,
                "value3": r2.value3,
            },
        )
        assert joined.op() == expected


def test_join_key_invalid(con):
    t1 = con.table("star1")
    t2 = con.table("star2")

    with pytest.raises(ExpressionError):
        t1.inner_join(t2, [("foo_id", "foo_id", "foo_id")])

    # it is working now
    t1.inner_join(t2, [(s.c("foo_id"), s.c("foo_id"))])


def test_join_invalid_refs(con):
    t1 = con.table("star1")
    t2 = con.table("star2")
    t3 = con.table("star3")

    predicate = t1.bar_id == t3.bar_id
    with pytest.raises(com.IntegrityError):
        t1.inner_join(t2, [predicate])


def test_join_invalid_expr_type(con):
    left = con.table("star1")
    invalid_right = left.foo_id
    join_key = ["bar_id"]

    with pytest.raises(TypeError):
        left.inner_join(invalid_right, join_key)


def test_join_non_boolean_expr(con):
    t1 = con.table("star1")
    t2 = con.table("star2")

    # oops
    predicate = t1.f * t2.value1
    with pytest.raises(ValidationError):
        t1.inner_join(t2, [predicate])


def test_unravel_compound_equijoin(table):
    t1 = ibis.table(
        [
            ("key1", "string"),
            ("key2", "string"),
            ("key3", "string"),
            ("value1", "double"),
        ],
        "foo_table",
    )

    t2 = ibis.table(
        [
            ("key1", "string"),
            ("key2", "string"),
            ("key3", "string"),
            ("value2", "double"),
        ],
        "bar_table",
    )

    p1 = t1.key1 == t2.key1
    p2 = t1.key2 == t2.key2
    p3 = t1.key3 == t2.key3

    joined = t1.inner_join(t2, [p1 & p2 & p3])
    with join_tables(t1, t2) as (r1, r2):
        expected = ops.JoinChain(
            first=r1,
            rest=[
                ops.JoinLink(
                    "inner",
                    r2,
                    [r1.key1 == r2.key1, r1.key2 == r2.key2, r1.key3 == r2.key3],
                )
            ],
            values={
                "key1": r1.key1,
                "key2": r1.key2,
                "key3": r1.key3,
                "value1": r1.value1,
                "value2": r2.value2,
            },
        )
        assert joined.op() == expected


def test_union(
    setops_table_foo,
    setops_table_bar,
    setops_table_baz,
    setops_relation_error_message,
):
    result = setops_table_foo.union(setops_table_bar)
    assert isinstance(result.op().parent, ops.Union)
    assert not result.op().parent.distinct

    result = setops_table_foo.union(setops_table_bar, distinct=True)
    assert result.op().parent.distinct

    with pytest.raises(RelationError, match=setops_relation_error_message):
        setops_table_foo.union(setops_table_baz)


def test_intersection(
    setops_table_foo,
    setops_table_bar,
    setops_table_baz,
    setops_relation_error_message,
):
    result = setops_table_foo.intersect(setops_table_bar)
    assert isinstance(result.op().parent, ops.Intersection)

    with pytest.raises(RelationError, match=setops_relation_error_message):
        setops_table_foo.intersect(setops_table_baz)


def test_difference(
    setops_table_foo,
    setops_table_bar,
    setops_table_baz,
    setops_relation_error_message,
):
    result = setops_table_foo.difference(setops_table_bar)
    assert isinstance(result.op().parent, ops.Difference)

    with pytest.raises(RelationError, match=setops_relation_error_message):
        setops_table_foo.difference(setops_table_baz)


def test_column_ref_on_projection_rename(con):
    region = con.table("tpch_region")
    nation = con.table("tpch_nation")
    customer = con.table("tpch_customer")

    joined = region.inner_join(
        nation, [region.r_regionkey == nation.n_regionkey]
    ).inner_join(customer, [customer.c_nationkey == nation.n_nationkey])

    proj_exprs = [
        customer,
        nation.n_name.name("nation"),
        region.r_name.name("region"),
    ]
    joined = joined.select(proj_exprs)

    metrics = [joined.c_acctbal.sum().name("metric")]

    # it works!
    joined.aggregate(metrics, by=["region"])


@pytest.fixture
def t1():
    return ibis.table(
        [("key1", "string"), ("key2", "string"), ("value1", "double")], "foo"
    )


@pytest.fixture
def t2():
    return ibis.table([("key1", "string"), ("key2", "string")], "bar")


def test_unresolved_existence_predicate(t1, t2):
    expr = (t1.key1 == t2.key1).any()
    assert isinstance(expr, Deferred)

    filtered = t2.filter(t1.key1 == t2.key1)
    subquery = ops.ExistsSubquery(filtered)
    expected = ops.Filter(parent=t1, predicates=[subquery])
    assert t1[expr].op() == expected

    filtered = t1.filter(t1.key1 == t2.key1)
    subquery = ops.ExistsSubquery(filtered)
    expected = ops.Filter(parent=t2, predicates=[subquery])
    assert t2[expr].op() == expected


def test_resolve_existence_predicate(t1, t2):
    expr = t1[(t1.key1 == t2.key1).any()]
    op = expr.op()
    assert isinstance(op, ops.Filter)

    pred = op.predicates[0].to_expr()
    assert isinstance(pred.op(), ops.ExistsSubquery)


def test_aggregate_metrics(table):
    functions = [
        lambda x: x.e.sum().name("esum"),
        lambda x: x.f.sum().name("fsum"),
    ]
    exprs = [table.e.sum().name("esum"), table.f.sum().name("fsum")]

    result = table.aggregate(functions[0])
    expected = table.aggregate(exprs[0])
    assert_equal(result, expected)

    result = table.aggregate(functions)
    expected = table.aggregate(exprs)
    assert_equal(result, expected)


def test_group_by_keys(table):
    m = table.mutate(foo=table.f * 2, bar=table.e / 2)

    expr = m.group_by(lambda x: x.foo).size()
    expected = m.group_by("foo").size()
    assert_equal(expr, expected)

    expr = m.group_by([lambda x: x.foo, lambda x: x.bar]).size()
    expected = m.group_by(["foo", "bar"]).size()
    assert_equal(expr, expected)


def test_having(table):
    m = table.mutate(foo=table.f * 2, bar=table.e / 2)
    expr = m.group_by("foo").having(lambda x: x.foo.sum() > 10).size()

    agg = ops.Aggregate(
        parent=m,
        groups={"foo": m.foo},
        metrics={"CountStar()": ops.CountStar(m), "Sum(foo)": ops.Sum(m.foo)},
    ).to_expr()
    filt = ops.Filter(
        parent=agg,
        predicates=[agg["Sum(foo)"] > 10],
    ).to_expr()
    proj = ops.Project(
        parent=filt,
        values={"foo": filt.foo, "CountStar()": filt["CountStar()"]},
    ).to_expr()

    assert expr.equals(proj)


def test_filter(table):
    m = table.mutate(foo=table.f * 2, bar=table.e / 2)

    result = m.filter(lambda x: x.foo > 10)
    result2 = m[lambda x: x.foo > 10]
    expected = m[m.foo > 10]

    assert_equal(result, expected)
    assert_equal(result2, expected)

    result = m.filter([lambda x: x.foo > 10, lambda x: x.bar < 0])
    expected = m.filter([m.foo > 10, m.bar < 0])
    assert_equal(result, expected)


def test_order_by2(table):
    m = table.mutate(foo=table.e + table.f)

    result = m.order_by(lambda x: -x.foo)
    expected = m.order_by(-m.foo)
    assert_equal(result, expected)

    result = m.order_by(lambda x: ibis.desc(x.foo))
    expected = m.order_by(ibis.desc("foo"))
    assert_equal(result, expected)

    result = m.order_by(ibis.desc(lambda x: x.foo))
    expected = m.order_by(ibis.desc("foo"))
    assert_equal(result, expected)

    result = m.order_by(ibis.asc(lambda x: x.foo))
    expected = m.order_by("foo")
    assert_equal(result, expected)


def test_projection2(table):
    m = table.mutate(foo=table.f * 2)

    def f(x):
        return (x.foo * 2).name("bar")

    result = m.select([f, "f"])
    result2 = m[f, "f"]
    expected = m.select([f(m), "f"])
    assert_equal(result, expected)
    assert_equal(result2, expected)


def test_mutate2(table):
    m = table.mutate(foo=table.f * 2)

    def g(x):
        return x.foo * 2

    def h(x):
        return x.bar * 2

    result = m.mutate(bar=g).mutate(baz=h)

    m2 = m.mutate(bar=g(m))
    expected = m2.mutate(baz=h(m2))

    assert_equal(result, expected)


def test_groupby_mutate(table):
    t = table

    g = t.group_by("g").order_by("f")
    expr = g.mutate(foo=lambda x: x.f.lag(), bar=lambda x: x.f.rank())
    expected = g.mutate(foo=t.f.lag(), bar=t.f.rank())

    assert_equal(expr, expected)


def test_groupby_projection(table):
    t = table

    g = t.group_by("g").order_by("f")
    expr = g.select([lambda x: x.f.lag().name("foo"), lambda x: x.f.rank().name("bar")])
    expected = g.select([t.f.lag().name("foo"), t.f.rank().name("bar")])

    assert_equal(expr, expected)


def test_pickle_table_expr():
    schema = [("time", "timestamp"), ("key", "string"), ("value", "double")]
    t0 = ibis.table(schema, name="t0")
    raw = pickle.dumps(t0, protocol=2)
    t1 = pickle.loads(raw)
    assert t1.equals(t0)


def test_pickle_table_node(table):
    n0 = table.op()
    assert_pickle_roundtrip(n0)


def test_pickle_projection_node(table):
    m = table.mutate(foo=table.f * 2)

    def f(x):
        return (x.foo * 2).name("bar")

    node = m.select([f, "f"]).op()

    assert_pickle_roundtrip(node)


def test_pickle_group_by(table):
    m = table.mutate(foo=table.f * 2, bar=table.e / 2)
    expr = m.group_by(lambda x: x.foo).size()
    node = expr.op()

    assert_pickle_roundtrip(node)


def test_pickle_asof_join():
    left = ibis.table([("time", "int32"), ("value", "double")])
    right = ibis.table([("time", "int32"), ("value2", "double")])
    joined = api.asof_join(left, right, "time")
    node = joined.op()

    assert_pickle_roundtrip(node)


def test_group_by_key_function():
    t = ibis.table([("a", "timestamp"), ("b", "string"), ("c", "double")])
    expr = t.group_by(new_key=lambda t: t.b.length()).aggregate(foo=t.c.mean())
    assert expr.columns == ["new_key", "foo"]


def test_group_by_no_keys():
    t = ibis.table([("a", "timestamp"), ("b", "string"), ("c", "double")])

    with pytest.raises(com.IbisInputError):
        t.group_by(s.startswith("x")).aggregate(foo=t.c.mean())


def test_unbound_table_name():
    t = ibis.table([("a", "timestamp")])
    name = t.op().name
    match = re.match(r"^unbound_table_\d+$", name)
    assert match is not None


class MyTable:
    a: int
    b: str
    c: list[float]


def test_unbound_table_using_class_definition():
    expected_schema = ibis.schema({"a": "int64", "b": "string", "c": "array<double>"})

    t1 = ibis.table(MyTable)
    t2 = ibis.table(MyTable, name="MyNamedTable")

    cases = {t1: "MyTable", t2: "MyNamedTable"}
    for t, name in cases.items():
        assert isinstance(t, ir.Table)
        assert isinstance(t.op(), ops.UnboundTable)
        assert t.schema() == expected_schema
        assert t.get_name() == name


def test_mutate_chain():
    one = ibis.table([("a", "string"), ("b", "string")], name="t")
    two = one.mutate(b=lambda t: t.b.fillna("Short Term"))
    three = two.mutate(a=lambda t: t.a.fillna("Short Term"))

    values = three.op().values
    assert isinstance(values["a"], ops.Coalesce)
    assert isinstance(values["b"], ops.Field)
    assert values["b"].rel == two.op()

    three_opt = simplify(three.op())
    assert three_opt == ops.Project(
        parent=one,
        values={
            "a": one.a.fillna("Short Term"),
            "b": one.b.fillna("Short Term"),
        },
    )


# TODO(kszucs): move this test case to ibis/tests/sql since it requires the
# sql backend to be executed
# def test_multiple_dbcon():
#     """Expr from multiple connections to same DB should be compatible."""
#     con1 = MockBackend()
#     con2 = MockBackend()

#     con1.table("alltypes").union(con2.table("alltypes")).execute()


# TODO(kszucs): move this test to ibis/tests/sql since it requires the
# sql backend to be executed
# def test_multiple_db_different_backends():
#     con1 = MockBackend()
#     con2 = MockAlchemyBackend()

#     backend1_table = con1.table("alltypes")
#     backend2_table = con2.table("alltypes")

#     expr = backend1_table.union(backend2_table)
#     with pytest.raises(com.IbisError, match="Multiple backends"):
#         expr.compile()


def test_merge_as_of_allows_overlapping_columns():
    # GH3295
    table = ibis.table(
        [
            ("field", "string"),
            ("value", "float64"),
            ("timestamp_received", "timestamp"),
        ],
        name="t",
    )

    signal_one = table[
        table["field"].contains("signal_one") & table["field"].contains("current")
    ]
    signal_one = signal_one[
        "value", "timestamp_received", "field"
    ]  # select columns we care about
    signal_one = signal_one.rename(current="value", signal_one="field")

    signal_two = table[
        table["field"].contains("signal_two") & table["field"].contains("voltage")
    ]
    signal_two = signal_two[
        "value", "timestamp_received", "field"
    ]  # select columns we care about
    signal_two = signal_two.rename(voltage="value", signal_two="field")

    merged = signal_one.asof_join(signal_two, "timestamp_received")
    assert merged.columns == [
        "current",
        "timestamp_received",
        "signal_one",
        "voltage",
        "timestamp_received_right",
        "signal_two",
    ]


def test_select_from_unambiguous_join_with_strings():
    # GH1387
    t = ibis.table([("a", "int64"), ("b", "string")])
    s = ibis.table([("b", "int64"), ("c", "string")])
    joined = t.left_join(s, [t.b == s.c])
    expr = joined[t, "c"]
    assert expr.columns == ["a", "b", "c"]


def test_filter_applied_to_join():
    # GH2437
    countries = ibis.table([("iso_alpha3", "string")])
    gdp = ibis.table([("country_code", "string"), ("year", "int64")])

    expr = countries.inner_join(
        gdp,
        predicates=[countries["iso_alpha3"] == gdp["country_code"]],
    ).filter(gdp["year"] == 2017)
    assert expr.columns == ["iso_alpha3", "country_code", "year"]


@pytest.mark.parametrize("how", ["inner", "left", "outer", "right"])
def test_join_lname_rname(how):
    left = ibis.table([("id", "int64"), ("first_name", "string")])
    right = ibis.table([("id", "int64"), ("last_name", "string")])
    method = getattr(left, f"{how}_join")

    expr = method(right)
    assert expr.columns == ["id", "first_name", "id_right", "last_name"]

    expr = method(right, rname="right_{name}")
    assert expr.columns == ["id", "first_name", "right_id", "last_name"]

    expr = method(right, lname="left_{name}", rname="")
    assert expr.columns == ["left_id", "first_name", "id", "last_name"]

    expr = method(right, rname="right_{name}", lname="left_{name}")
    assert expr.columns == ["left_id", "first_name", "right_id", "last_name"]


def test_join_lname_rname_still_collide():
    t1 = ibis.table({"id": "int64", "col1": "int64", "col2": "int64"})
    t2 = ibis.table({"id": "int64", "col1": "int64", "col2": "int64"})
    t3 = ibis.table({"id": "int64", "col1": "int64", "col2": "int64"})

    with pytest.raises(com.IntegrityError):
        t1.left_join(t2, "id").left_join(t3, "id")._finish()

    # assert "`['col1_right', 'col2_right', 'id_right']`" in str(rec.value)
    # assert "`lname='', rname='{name}_right'`" in str(rec.value)


def test_drop():
    t = ibis.table(dict.fromkeys("abcd", "int"))

    assert t.drop() is t

    res = t.drop("a")
    assert res.equals(t.select("b", "c", "d"))

    res = t.drop("a", "b")
    assert res.equals(t.select("c", "d"))

    assert res.equals(t.select("c", "d"))

    assert res.equals(t.drop(s.matches("a|b")))

    res = t.drop(_.a)
    assert res.equals(t.select("b", "c", "d"))

    res = t.drop(_.a, _.b)
    assert res.equals(t.select("c", "d"))

    res = t.drop(_.a, "b")
    assert res.equals(t.select("c", "d"))

    with pytest.raises(KeyError):
        t.drop("e")


def test_python_table_ambiguous():
    with pytest.raises(NotImplementedError):
        ibis.memtable(
            [(1,)],
            schema=ibis.schema(dict(a="int8")),
            columns=["a"],
        )


def test_memtable_filter():
    # Mostly just a smoketest, this used to error on construction
    t = ibis.memtable([(1, 2), (3, 4), (5, 6)], columns=["x", "y"])
    expr = t.filter(t.x > 1)
    assert expr.columns == ["x", "y"]


def test_default_backend_with_unbound_table():
    t = ibis.table(dict(a="int"), name="t")
    expr = t.a.sum()

    with pytest.raises(
        com.IbisError,
        match="Expression contains unbound tables",
    ):
        assert expr.execute()


def test_numpy_ufuncs_dont_cast_tables():
    t = ibis.table(dict.fromkeys("abcd", "int"))
    for arg in [np.int64(1), np.array([1, 2, 3])]:
        for left, right in [(t, arg), (arg, t)]:
            with pytest.raises(TypeError):
                left + right


def test_array_string_compare():
    t = ibis.table(schema=dict(by="string", words="array<string>"), name="t")
    expr = t[t.by == "foo"].mutate(words=_.words.unnest()).filter(_.words == "the")
    assert expr is not None


@pytest.mark.parametrize("value", [True, False])
def test_filter_with_literal(value):
    t = ibis.table(dict(a="string"))
    filt = t.filter(ibis.literal(value))
    assert filt.op() == ops.Filter(parent=t, predicates=[ibis.literal(value)])

    # ints are invalid predicates
    int_val = ibis.literal(int(value))
    with pytest.raises(ValidationError):
        t.filter(int_val)


def test_cast():
    t = ibis.table(dict(a="int", b="string", c="float"), name="t")

    assert t.cast({"a": "string"}).equals(t.mutate(a=t.a.cast("string")))

    with pytest.raises(
        com.IbisError, match="fields that are not in the table: .+'d'.+"
    ):
        t.cast({"d": "array<int>"}).equals(t.select())

    assert t.cast(ibis.schema({"a": "string", "b": "int"})).equals(
        t.mutate(a=t.a.cast("string"), b=t.b.cast("int"))
    )
    assert t.cast([("a", "string"), ("b", "float")]).equals(
        t.mutate(a=t.a.cast("string"), b=t.b.cast("float"))
    )


def test_pivot_longer():
    diamonds = ibis.table(
        {
            "carat": "float64",
            "cut": "string",
            "color": "string",
            "clarity": "string",
            "depth": "float64",
            "table": "float64",
            "price": "int64",
            "x": "float64",
            "y": "float64",
            "z": "float64",
        },
        name="diamonds",
    )
    res = diamonds.pivot_longer(s.c("x", "y", "z"), names_to="pos", values_to="xyz")
    assert res.schema().names == (
        "carat",
        "cut",
        "color",
        "clarity",
        "depth",
        "table",
        "price",
        "pos",
        "xyz",
    )


def test_pivot_longer_strip_prefix():
    t = ibis.table(
        dict(artist="string", track="string", wk1="int", wk2="int", wk3="int")
    )
    expr = t.pivot_longer(
        s.startswith("wk"),
        names_to="week",
        names_pattern=r"wk(.+)",
        names_transform=int,
        values_to="rank",
        values_transform=_.cast("int"),
    )
    schema = ibis.schema(dict(artist="string", track="string", week="int8", rank="int"))
    assert expr.schema() == schema


def test_pivot_longer_pluck_regex():
    t = ibis.table(
        dict(artist="string", track="string", x_wk1="int", x_wk2="int", x_wk3="int")
    )
    expr = t.pivot_longer(
        s.matches("^.+wk.$"),
        names_to=["other_var", "week"],
        names_pattern=r"(.)_wk(\d)",
        names_transform=dict(other_var=str.upper, week=int),
        values_to="rank",
        values_transform=_.cast("int"),
    )
    schema = ibis.schema(
        dict(
            artist="string", track="string", other_var="string", week="int8", rank="int"
        )
    )
    assert expr.schema() == schema


def test_pivot_longer_no_match():
    t = ibis.table(
        dict(artist="string", track="string", x_wk1="int", x_wk2="int", x_wk3="int")
    )
    with pytest.raises(
        com.IbisInputError, match="Selector returned no columns to pivot on"
    ):
        t.pivot_longer(
            s.matches("foo"),
            names_to=["other_var", "week"],
            names_pattern=r"(.)_wk(\d)",
            names_transform=dict(other_var=str.upper, week=int),
            values_to="rank",
            values_transform=_.cast("int"),
        )


def test_pivot_wider():
    fish = ibis.table({"fish": "int", "station": "string", "seen": "int"}, name="fish")
    res = fish.pivot_wider(
        names=["Release", "Lisbon"], names_from="station", values_from="seen"
    )
    assert res.schema().names == ("fish", "Release", "Lisbon")
    with pytest.raises(com.IbisInputError, match="Columns .+ are not present in"):
        fish.pivot_wider(names=["Release", "Lisbon"], values_from="seen")


def test_invalid_deferred():
    t = ibis.table(dict(value="int", lagged_value="int"), name="t")

    with pytest.raises(ValidationError):
        ops.Greatest((t.value, ibis._.lagged_value))


@pytest.mark.parametrize("keep", ["last", None])
def test_invalid_distinct(keep):
    t = ibis.table(dict(a="int"), name="t")
    with pytest.raises(com.IbisError, match="Only keep='first'"):
        t.distinct(keep=keep)


def test_invalid_keep_distinct():
    t = ibis.table(dict(a="int", b="string"), name="t")
    with pytest.raises(com.IbisError, match="Invalid value for `keep`:"):
        t.distinct(on="a", keep="invalid")


def test_invalid_distinct_empty_key():
    t = ibis.table(dict(a="int", b="string"), name="t")
    with pytest.raises(com.IbisInputError):
        t.distinct(on="c", keep="first")


def test_unbind_with_namespace():
    schema = ibis.schema({"a": "int"})
    ns = ops.Namespace(database="db", schema="sch")

    t_op = ops.DatabaseTable(name="t", schema=schema, source=None, namespace=ns)
    t = t_op.to_expr()
    s = t.unbind()

    expected = ops.UnboundTable(name="t", schema=schema, namespace=ns).to_expr()

    assert s.op() == expected.op()
    assert s.equals(expected)



from __future__ import annotations

import pandas as pd
import pandas.testing as tm
import pytest
from pytest import param

import ibis

cc = pytest.importorskip("clickhouse_connect")


@pytest.fixture(scope="module")
def diamonds(con):
    return con.table("diamonds")


@pytest.fixture(scope="module")
def batting(con):
    return con.table("batting")


@pytest.fixture(scope="module")
def awards_players(con):
    return con.table("awards_players")


@pytest.fixture(scope="module")
def time_left(con):
    return con.table("time_df1")


@pytest.fixture(scope="module")
def time_right(con):
    return con.table("time_df2")


def test_timestamp_extract_field(alltypes, snapshot):
    t = alltypes.timestamp_col
    expr = alltypes[
        t.year().name("year"),
        t.month().name("month"),
        t.day().name("day"),
        t.hour().name("hour"),
        t.minute().name("minute"),
        t.second().name("second"),
    ]

    result = ibis.clickhouse.compile(expr)

    snapshot.assert_match(result, "out.sql")


def test_isin_notin_in_select(alltypes, snapshot):
    values = ["foo", "bar"]
    filtered = alltypes[alltypes.string_col.isin(values)]
    result = ibis.clickhouse.compile(filtered)
    snapshot.assert_match(result, "out1.sql")

    filtered = alltypes[alltypes.string_col.notin(values)]
    result = ibis.clickhouse.compile(filtered)
    snapshot.assert_match(result, "out2.sql")


def test_head(alltypes):
    result = alltypes.head().execute()
    expected = alltypes.limit(5).execute()
    tm.assert_frame_equal(result, expected)


def test_limit_offset(alltypes):
    expected = alltypes.execute()

    tm.assert_frame_equal(alltypes.limit(4).execute(), expected.head(4))
    tm.assert_frame_equal(alltypes.limit(8).execute(), expected.head(8))
    tm.assert_frame_equal(
        alltypes.limit(4, offset=2).execute(),
        expected.iloc[2:6].reset_index(drop=True),
    )


def test_subquery(alltypes, df):
    t = alltypes

    expr = t.mutate(d=t.double_col).limit(100).group_by("string_col").size()
    result = expr.execute()

    result = result.sort_values("string_col").reset_index(drop=True)
    expected = (
        df.assign(d=df.double_col.fillna(0))
        .head(100)
        .groupby("string_col")
        .string_col.count()
        .rename("CountStar()")
        .reset_index()
        .sort_values("string_col")
        .reset_index(drop=True)
    )

    tm.assert_frame_equal(result, expected)


def test_simple_scalar_aggregates(alltypes, snapshot):
    # Things like table.column.{sum, mean, ...}()
    table = alltypes

    expr = table[table.int_col > 0].float_col.sum()

    sql_query = ibis.clickhouse.compile(expr)
    snapshot.assert_match(sql_query, "out.sql")


# def test_scalar_aggregates_multiple_tables(alltypes):
#     # #740
#     table = ibis.table([('flag', 'string'),
#                         ('value', 'double')],
#                        'tbl')

#     flagged = table[table.flag == '1']
#     unflagged = table[table.flag == '0']

#     expr = flagged.value.mean() / unflagged.value.mean() - 1

#     result = ibis.clickhouse.compile(expr)
#     expected = """\
# SELECT (t0.`mean` / t1.`mean`) - 1 AS `tmp`
# FROM (
#   SELECT avg(`value`) AS `mean`
#   FROM tbl
#   WHERE `flag` = '1'
# ) t0
#   CROSS JOIN (
#     SELECT avg(`value`) AS `mean`
#     FROM tbl
#     WHERE `flag` = '0'
#   ) t1"""
#     assert result == expected

#     fv = flagged.value
#     uv = unflagged.value

#     expr = (fv.mean() / fv.sum()) - (uv.mean() / uv.sum())
#     result = ibis.clickhouse.compile(expr)
#     expected = """\
# SELECT t0.`tmp` - t1.`tmp` AS `tmp`
# FROM (
#   SELECT avg(`value`) / sum(`value`) AS `tmp`
#   FROM tbl
#   WHERE `flag` = '1'
# ) t0
#   CROSS JOIN (
#     SELECT avg(`value`) / sum(`value`) AS `tmp`
#     FROM tbl
#     WHERE `flag` = '0'
#   ) t1"""
#     assert result == expected


def test_table_column_unbox(alltypes, snapshot):
    m = alltypes.float_col.sum().name("total")
    agged = alltypes[alltypes.int_col > 0].group_by("string_col").aggregate([m])
    expr = agged.string_col

    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_complex_array_expr_projection(alltypes, snapshot):
    # May require finding the base table and forming a projection.
    expr = (
        alltypes.group_by("string_col")
        .aggregate([alltypes.count().name("count")])
        .string_col.cast("double")
    )

    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "expr",
    [param(ibis.now(), id="now"), param(ibis.literal(1) + ibis.literal(2), id="add")],
)
def test_scalar_exprs_no_table_refs(expr, snapshot):
    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_isnull_case_expr_rewrite_failure(alltypes, snapshot):
    # #172, case expression that was not being properly converted into an
    # aggregation
    reduction = alltypes.string_col.isnull().ifelse(1, 0).sum()

    result = ibis.clickhouse.compile(reduction)
    snapshot.assert_match(result, "out.sql")


def test_physical_table_reference_translate(alltypes, snapshot):
    # If an expression's table leaves all reference database tables, verify
    # we translate correctlys
    result = ibis.clickhouse.compile(alltypes)
    snapshot.assert_match(result, "out.sql")


def test_non_equijoin(alltypes):
    t = alltypes.limit(100)
    t2 = t.view()
    expr = t.join(t2, t.tinyint_col < t2.timestamp_col.minute()).count()

    # compilation should pass
    expr.compile()

    # while execution should fail since clickhouse doesn't support non-equijoin
    with pytest.raises(Exception, match="Unsupported JOIN ON conditions"):
        expr.execute()


@pytest.mark.parametrize(
    "join_type",
    ["any_inner_join", "inner_join", "any_left_join", "left_join"],
)
@pytest.mark.parametrize(
    ("left_key", "right_key"), [("playerID", "playerID"), ("playerID", "awardID")]
)
def test_simple_joins(
    batting, awards_players, join_type, left_key, right_key, snapshot
):
    t1, t2 = batting, awards_players
    pred = [t1[left_key] == t2[right_key]]
    expr = getattr(t1, join_type)(t2, pred)[[t1]]

    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_self_reference_simple(con, alltypes, snapshot):
    expr = alltypes.view()
    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")
    assert len(con.execute(expr))


def test_join_self_reference(con, alltypes, snapshot):
    t1 = alltypes
    t2 = t1.view()
    expr = t1.inner_join(t2, ["id"])[[t1]]

    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")
    assert len(con.execute(expr))


def test_where_simple_comparisons(con, alltypes, snapshot):
    t1 = alltypes
    expr = t1.filter([t1.float_col > 0, t1.int_col < t1.float_col * 2])

    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")
    assert len(con.execute(expr))


def test_where_with_between(con, alltypes, snapshot):
    t = alltypes

    expr = t.filter([t.int_col > 0, t.float_col.between(0, 1)])
    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")
    con.execute(expr)


def test_ifelse_use_if(con, alltypes, snapshot):
    expr = ibis.ifelse(alltypes.float_col > 0, alltypes.int_col, alltypes.bigint_col)

    result = expr.compile()
    snapshot.assert_match(result, "out.sql")
    con.execute(expr)


def test_filter_predicates(diamonds):
    predicates = [
        lambda x: x.color.lower().like("%de%"),
        # lambda x: x.color.lower().contains('de'),
        lambda x: x.color.lower().rlike(".*ge.*"),
    ]

    expr = diamonds
    for pred in predicates:
        expr = expr[pred(expr)].select(expr)

    expr.execute()


def test_where_with_timestamp(snapshot):
    t = ibis.table(
        [("uuid", "string"), ("ts", "timestamp"), ("search_level", "int64")],
        name="t",
    )
    expr = t.group_by(t.uuid).aggregate(min_date=t.ts.min(where=t.search_level == 1))
    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_timestamp_scalar_in_filter(alltypes):
    table = alltypes

    expr = table.filter(
        [
            table.timestamp_col
            < (ibis.timestamp("2010-01-01") + ibis.interval(weeks=3)),
            table.timestamp_col < (ibis.now() + ibis.interval(days=10)),
        ]
    ).count()
    expr.execute()


def test_named_from_filter_groupby(snapshot):
    t = ibis.table([("key", "string"), ("value", "double")], name="t0")
    gb = t.filter(t.value == 42).group_by(t.key)
    sum_expr = lambda t: (t.value + 1 + 2 + 3).sum()
    expr = gb.aggregate(abc=sum_expr)
    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out1.sql")

    expr = gb.aggregate(foo=sum_expr)
    result = ibis.clickhouse.compile(expr)
    snapshot.assert_match(result, "out2.sql")


def test_join_with_external_table_errors(alltypes):
    external_table = ibis.table(
        [("a", "string"), ("b", "int64"), ("c", "string")], name="external"
    )

    alltypes = alltypes.mutate(b=alltypes.tinyint_col)
    expr = alltypes.inner_join(external_table, ["b"])[
        external_table.a, external_table.c, alltypes.id
    ]

    with pytest.raises(cc.driver.exceptions.DatabaseError):
        expr.execute()

    with pytest.raises(TypeError):
        expr.execute(external_tables={"external": []})


def test_join_with_external_table(alltypes, df):
    external_df = pd.DataFrame(
        [("alpha", 1, "first"), ("beta", 2, "second"), ("gamma", 3, "third")],
        columns=["a", "b", "c"],
    )
    external_df["b"] = external_df["b"].astype("int8")

    external_table = ibis.table(
        [("a", "string"), ("b", "int64"), ("c", "string")], name="external"
    )

    alltypes = alltypes.mutate(b=alltypes.tinyint_col)
    expr = alltypes.inner_join(external_table, ["b"])[
        external_table.a, external_table.c, alltypes.id
    ]

    result = expr.execute(external_tables={"external": external_df})
    expected = df.assign(b=df.tinyint_col).merge(external_df, on="b")[["a", "c", "id"]]

    result = result.sort_values("id").reset_index(drop=True)
    expected = expected.sort_values("id").reset_index(drop=True)

    tm.assert_frame_equal(result, expected, check_column_type=False)


def test_asof_join(time_left, time_right):
    expr = time_left.asof_join(
        time_right,
        on=time_left["time"] >= time_right["time"],
        predicates=[
            time_left["key"] == time_right["key"],
        ],
    ).drop("time_right")
    result = expr.execute()
    expected = pd.merge_asof(
        time_left.execute(), time_right.execute(), on="time", suffixes=("", "_right")
    )
    tm.assert_frame_equal(
        result[expected.columns].fillna(pd.NA), expected.fillna(pd.NA)
    )


def test_count_name(snapshot):
    t = ibis.table(dict(a="string", b="bool"), name="t")

    expr = t.group_by(t.a).agg(
        A=t.count(where=~t.b).fillna(0), B=t.count(where=t.b).fillna(0)
    )

    snapshot.assert_match(str(ibis.to_sql(expr, dialect="clickhouse")), "out.sql")


def test_array_join_in_subquery(snapshot):
    node_view = ibis.table(dict(id="int64"), name="node_view")
    way_view = ibis.table(dict(ids="array<!int64>"), name="way_view")

    expr = node_view.id.isin(way_view.ids.unnest())

    out = ibis.clickhouse.compile(expr)
    snapshot.assert_match(out, "out.sql")


def test_complex_join(snapshot):
    t1 = ibis.table({"a": "int", "b": "int"}, name="s")
    t2 = ibis.table({"c": "int", "d": "int"}, name="t")
    t3 = t1.join(t2, t1.a == t2.c)
    q = t3.mutate(e=t3.c / (t3.a - t3.b))
    out = ibis.clickhouse.compile(q)
    snapshot.assert_match(out, "out.sql")



import ibis


functional_alltypes = ibis.table(
    name="functional_alltypes",
    schema={
        "id": "int32",
        "bool_col": "boolean",
        "tinyint_col": "int8",
        "smallint_col": "int16",
        "int_col": "int32",
        "bigint_col": "int64",
        "float_col": "float32",
        "double_col": "float64",
        "date_string_col": "string",
        "string_col": "string",
        "timestamp_col": "timestamp",
        "year": "int32",
        "month": "int32",
    },
)
lit = ibis.literal(0)
f = functional_alltypes.filter(functional_alltypes.int_col > lit)
f1 = functional_alltypes.filter(functional_alltypes.int_col <= lit)

result = f.select(
    f.string_col.name("key"), f.float_col.cast("float64").name("value")
).union(
    f1.select(f1.string_col.name("key"), f1.double_col.name("value")), distinct=True
)



from __future__ import annotations

import sqlglot.expressions as sge
from public import public

import ibis.common.exceptions as com
import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.base.sqlglot.compiler import ALL_OPERATIONS
from ibis.backends.base.sqlglot.datatypes import RisingWaveType
from ibis.backends.base.sqlglot.dialects import RisingWave
from ibis.backends.postgres.compiler import PostgresCompiler


@public
class RisingwaveCompiler(PostgresCompiler):
    __slots__ = ()

    dialect = RisingWave
    type_mapper = RisingWaveType

    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.DateFromYMD,
            ops.Mode,
            *(
                op
                for op in ALL_OPERATIONS
                if issubclass(
                    op, (ops.GeoSpatialUnOp, ops.GeoSpatialBinOp, ops.GeoUnaryUnion)
                )
            ),
        )
    )

    SIMPLE_OPS = {
        ops.First: "first_value",
        ops.Last: "last_value",
    }

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                "RisingWave only implements `pop` correlation coefficient"
            )
        return super().visit_Correlation(
            op, left=left, right=right, how=how, where=where
        )

    def visit_TimestampTruncate(self, op, *, arg, unit):
        unit_mapping = {
            "Y": "year",
            "Q": "quarter",
            "M": "month",
            "W": "week",
            "D": "day",
            "h": "hour",
            "m": "minute",
            "s": "second",
            "ms": "milliseconds",
            "us": "microseconds",
        }

        if (unit := unit_mapping.get(unit.short)) is None:
            raise com.UnsupportedOperationError(f"Unsupported truncate unit {unit}")

        return self.f.date_trunc(unit, arg)

    visit_TimeTruncate = visit_DateTruncate = visit_TimestampTruncate

    def visit_IntervalFromInteger(self, op, *, arg, unit):
        if op.arg.shape == ds.scalar:
            return sge.Interval(this=arg, unit=self.v[unit.name])
        elif op.arg.shape == ds.columnar:
            return arg * sge.Interval(this=sge.convert(1), unit=self.v[unit.name])
        else:
            raise ValueError("Invalid shape for converting to interval")

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            return self.cast("".join(map(r"\x{:0>2x}".format, value)), dt.binary)
        elif dtype.is_date():
            return self.cast(value.isoformat(), dtype)
        elif dtype.is_json():
            return sge.convert(str(value))
        return None



import ibis


star1 = ibis.table(
    name="star1",
    schema={"c": "int32", "f": "float64", "foo_id": "string", "bar_id": "string"},
)
star2 = ibis.table(
    name="star2", schema={"foo_id": "string", "value1": "float64", "value3": "float64"}
)
joinchain = star1.inner_join(star2, star1.foo_id == star2.foo_id).select(
    star1.c, star1.f, star1.foo_id, star1.bar_id, (star1.f - star2.value1).name("diff")
)

result = joinchain.filter(joinchain.diff > 1)



from __future__ import annotations

import contextlib
import sys
import tempfile
from html import escape

import graphviz as gv

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
from ibis.common.graph import Graph


def get_type(node):
    with contextlib.suppress(AttributeError, NotImplementedError):
        return escape(str(node.dtype))

    try:
        schema = node.schema
    except (AttributeError, NotImplementedError):
        # TODO(kszucs): this branch should be removed
        try:
            # As a last resort try get the name of the output_type class
            return node.output_type.__name__
        except (AttributeError, NotImplementedError):
            return "\u2205"  # empty set character
    except com.IbisError:
        assert isinstance(node, ops.Join)
        left_table_name = getattr(node.left, "name", None) or ops.genname()
        left_schema = node.left.schema
        right_table_name = getattr(node.right, "name", None) or ops.genname()
        right_schema = node.right.schema
        pairs = [
            (f"{left_table_name}.{left_column}", type)
            for left_column, type in left_schema.items()
        ] + [
            (f"{right_table_name}.{right_column}", type)
            for right_column, type in right_schema.items()
        ]
        schema = ibis.schema(pairs)

    return '<BR ALIGN="LEFT" />' + '<BR ALIGN="LEFT" />'.join(
        f"<I>{escape(name)}</I>: {escape(str(type))}"
        for name, type in zip(schema.names, schema.types)
    )


def get_label(node):
    typename = get_type(node)  # Already an escaped string
    name = type(node).__name__
    nodename = (
        node.name
        if isinstance(
            node,
            (
                ops.Literal,
                ops.Field,
                ops.Alias,
                ops.PhysicalTable,
                ops.window.RangeWindowFrame,
            ),
        )
        else None
    )
    if nodename is not None:
        # [TODO] Don't show nodename because it's too long and ruins the image
        if isinstance(node, ops.window.RangeWindowFrame):
            label_fmt = "<<B>{}</B>>"
            label = label_fmt.format(escape(name))
        else:
            if isinstance(node, ops.Relation):
                label_fmt = "<<I>{}</I>: <B>{}</B>{}>"
            else:
                label_fmt = '<<I>{}</I>: <B>{}</B><BR ALIGN="LEFT" />:: {}>'
            # typename is already escaped
            label = label_fmt.format(escape(nodename), escape(name), typename)
    else:
        if isinstance(node, ops.Relation):
            label_fmt = "<<B>{}</B>{}>"
        else:
            label_fmt = '<<B>{}</B><BR ALIGN="LEFT" />:: {}>'
        label = label_fmt.format(escape(name), typename)
    return label


DEFAULT_NODE_ATTRS = {"shape": "box", "fontname": "Deja Vu Sans Mono"}
DEFAULT_EDGE_ATTRS = {"fontname": "Deja Vu Sans Mono"}


def to_graph(expr, node_attr=None, edge_attr=None, label_edges: bool = False):
    graph = Graph.from_bfs(expr.op(), filter=ops.Node)

    g = gv.Digraph(
        node_attr=node_attr or DEFAULT_NODE_ATTRS,
        edge_attr=edge_attr or DEFAULT_EDGE_ATTRS,
    )

    g.attr(rankdir="BT")

    seen = set()
    edges = set()

    for v, us in graph.items():
        vhash = str(hash(v))
        if v not in seen:
            g.node(vhash, label=get_label(v))
            seen.add(v)

        for u in us:
            uhash = str(hash(u))
            if u not in seen:
                g.node(uhash, label=get_label(u))
                seen.add(u)
            if (edge := (u, v)) not in edges:
                if not label_edges:
                    label = None
                else:
                    for name, arg in zip(v.argnames, v.args):
                        if isinstance(arg, tuple) and u in arg:
                            index = arg.index(u)
                            name = f"{name}[{index}]"
                            break
                        elif arg == u:
                            break
                    else:
                        name = None
                    label = f"<.{name}>"

                g.edge(uhash, vhash, label=label)
                edges.add(edge)
    return g


def draw(graph, path=None, format="png", verbose: bool = False):
    if verbose:
        print(graph.source, file=sys.stderr)  # noqa: T201

    piped_source = graph.pipe(format=format)

    if path is None:
        with tempfile.NamedTemporaryFile(
            delete=False, suffix=f".{format}", mode="wb"
        ) as f:
            f.write(piped_source)
        return f.name
    else:
        with open(path, mode="wb") as f:
            f.write(piped_source)
        return path


if __name__ == "__main__":
    from argparse import ArgumentParser

    from ibis import _

    p = ArgumentParser(
        description="Render a GraphViz SVG of an example ibis expression."
    )

    p.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Print GraphViz DOT code to stderr.",
    )
    p.add_argument(
        "-l",
        "--label-edges",
        action="store_true",
        help="Show operation inputs as edge labels.",
    )

    args = p.parse_args()

    left = ibis.table(dict(a="int64", b="string"), name="left")
    right = ibis.table(dict(b="string", c="int64", d="string"), name="right")
    expr = (
        left.inner_join(right, "b")
        .select(left.a, b=right.c, c=right.d)
        .filter((_.a + _.b * 2 * _.b / _.b**3 > 4) & (_.b > 5))
        .group_by(_.c)
        .having(_.a.mean() > 0.0)
        .aggregate(a_mean=_.a.mean(), b_sum=_.b.sum())
        .mutate(
            arrays=ibis.array([1, 2, 3]),
            maps=ibis.map({"a": 1, "b": 2}),
            structs=ibis.struct({"a": [1, 2, 3], "b": {"c": 1, "d": 2}}),
        )
    )
    expr.visualize(verbose=args.verbose > 0, label_edges=args.label_edges)



from __future__ import annotations

import pytest
from pytest import param

import ibis.expr.types as ir
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.errors import PyDruidProgrammingError


def test_backend_name(backend):
    # backend is the TestConf for the backend
    assert backend.api.name == backend.name()


@pytest.mark.notyet(["druid"], raises=PyDruidProgrammingError)
def test_version(backend):
    assert isinstance(backend.api.version, str)


# 1. `current_database` returns '.', but isn't listed in list_databases()
@pytest.mark.never(
    ["polars", "dask", "exasol", "pandas", "druid", "oracle", "bigquery"],
    reason="backend does not support databases",
    raises=AttributeError,
)
@pytest.mark.notimpl(
    ["datafusion"],
    raises=NotImplementedError,
    reason="current_database isn't implemented",
)
def test_database_consistency(backend, con):
    databases = con.list_databases()
    assert isinstance(databases, list)
    assert len(databases) >= 1
    assert all(isinstance(database, str) for database in databases)

    # every backend has a different set of databases, not testing the
    # exact names for now
    current_database = con.current_database
    assert isinstance(current_database, str)
    if backend.name() == "snowflake":
        assert current_database.upper() in databases
    else:
        assert current_database in databases


def test_list_tables(con):
    tables = con.list_tables()
    assert isinstance(tables, list)
    # only table that is guaranteed to be in all backends
    key = "functional_alltypes"
    assert key in tables or key.upper() in tables
    assert all(isinstance(table, str) for table in tables)


def test_tables_accessor_mapping(con):
    if con.name == "snowflake":
        pytest.skip("snowflake sometimes counts more tables than are around")

    name = "functional_alltypes"

    assert isinstance(con.tables[name], ir.Table)

    with pytest.raises(KeyError, match="doesnt_exist"):
        con.tables["doesnt_exist"]

    # temporary might pop into existence in parallel test runs, in between the
    # first `list_tables` call and the second, so we check that there's a
    # non-empty intersection
    assert TEST_TABLES.keys() & set(map(str.lower, con.list_tables()))
    assert TEST_TABLES.keys() & set(map(str.lower, con.tables))


def test_tables_accessor_getattr(con):
    name = "functional_alltypes"
    assert isinstance(getattr(con.tables, name), ir.Table)

    with pytest.raises(AttributeError, match="doesnt_exist"):
        con.tables.doesnt_exist  # noqa: B018

    # Underscore/double-underscore attributes are never available, since many
    # python apis expect checking for the absence of these to be cheap.
    with pytest.raises(AttributeError, match="_private_attr"):
        con.tables._private_attr  # noqa: B018


def test_tables_accessor_tab_completion(con):
    name = "functional_alltypes"
    attrs = dir(con.tables)
    assert name in attrs
    assert "keys" in attrs  # type methods also present

    keys = con.tables._ipython_key_completions_()
    assert name in keys


def test_tables_accessor_repr(con):
    name = "functional_alltypes"
    result = repr(con.tables)
    assert f"- {name}" in result


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t.limit(5).limit(10), id="small_big"),
        param(lambda t: t.limit(10).limit(5), id="big_small"),
    ],
)
def test_limit_chain(alltypes, expr_fn):
    expr = expr_fn(alltypes)
    result = expr.execute()
    assert len(result) == 5


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t, id="alltypes table"),
        param(lambda t: t.join(t.view(), [("id", "int_col")]), id="self join"),
    ],
)
def test_unbind(alltypes, expr_fn):
    expr = expr_fn(alltypes)
    assert expr.unbind() != expr
    assert expr.unbind().schema() == expr.schema()

    assert "Unbound" not in repr(expr)
    assert "Unbound" in repr(expr.unbind())



from __future__ import annotations

import decimal
import enum
import json
from collections import OrderedDict
from datetime import date, datetime, timedelta

import numpy as np
import pandas as pd
import pytest
import pytz
from packaging.version import parse as vparse

import ibis.expr.datatypes as dt


class Foo(enum.Enum):
    a = 1
    b = 2


@pytest.mark.parametrize(
    ("value", "expected_dtype"),
    [
        (None, dt.null),
        (False, dt.boolean),
        (True, dt.boolean),
        ("foo", dt.string),
        (b"fooblob", dt.binary),
        (date.today(), dt.date),
        (datetime.now(), dt.timestamp),
        (timedelta(days=3), dt.Interval(unit="D")),
        (pd.Timedelta("5 hours"), dt.Interval(unit="h")),
        (pd.Timedelta("7 minutes"), dt.Interval(unit="m")),
        (timedelta(seconds=9), dt.Interval(unit="s")),
        (pd.Timedelta("11 milliseconds"), dt.Interval(unit="ms")),
        (timedelta(microseconds=15), dt.Interval(unit="us")),
        (pd.Timedelta("17 nanoseconds"), dt.Interval(unit="ns")),
        # numeric types
        (5, dt.int8),
        (5, dt.int8),
        (127, dt.int8),
        (128, dt.int16),
        (32767, dt.int16),
        (32768, dt.int32),
        (2147483647, dt.int32),
        (2147483648, dt.int64),
        (-5, dt.int8),
        (-128, dt.int8),
        (-129, dt.int16),
        (-32769, dt.int32),
        (-2147483649, dt.int64),
        (1.5, dt.double),
        (decimal.Decimal(1.5), dt.decimal),
        # parametric types
        (list("abc"), dt.Array(dt.string)),
        (set("abc"), dt.Array(dt.string)),
        ({1, 5, 6}, dt.Array(dt.int8)),
        (frozenset(list("abc")), dt.Array(dt.string)),
        ([1, 2, 3], dt.Array(dt.int8)),
        ([1, 128], dt.Array(dt.int16)),
        ([1, 128, 32768], dt.Array(dt.int32)),
        ([1, 128, 32768, 2147483648], dt.Array(dt.int64)),
        ({"a": 1, "b": 2, "c": 3}, dt.Map(dt.string, dt.int8)),
        ({1: 2, 3: 4, 5: 6}, dt.Map(dt.int8, dt.int8)),
        (
            {"a": [1.0, 2.0], "b": [], "c": [3.0]},
            dt.Map(dt.string, dt.Array(dt.double)),
        ),
        (
            OrderedDict(
                [
                    ("a", 1),
                    ("b", list("abc")),
                    ("c", OrderedDict([("foo", [1.0, 2.0])])),
                ]
            ),
            dt.Struct.from_tuples(
                [
                    ("a", dt.int8),
                    ("b", dt.Array(dt.string)),
                    (
                        "c",
                        dt.Struct.from_tuples([("foo", dt.Array(dt.double))]),
                    ),
                ]
            ),
        ),
        (Foo.a, dt.Enum()),
        # numpy types
        (np.int8(5), dt.int8),
        (np.int16(-1), dt.int16),
        (np.int32(2), dt.int32),
        (np.int64(-5), dt.int64),
        (np.uint8(5), dt.uint8),
        (np.uint16(50), dt.uint16),
        (np.uint32(500), dt.uint32),
        (np.uint64(5000), dt.uint64),
        (np.float32(5.5), dt.float32),
        (np.float64(5.55), dt.float64),
        (np.bool_(True), dt.boolean),
        (np.bool_(False), dt.boolean),
        # pandas types
        (
            pd.Timestamp("2015-01-01 12:00:00", tz="US/Eastern"),
            dt.Timestamp("US/Eastern"),
        ),
    ],
)
def test_infer_dtype(value, expected_dtype):
    assert dt.infer(value) == expected_dtype


def test_infer_mixed_type_fails():
    data = [1, "a"]
    with pytest.raises(TypeError):
        dt.infer(data)


def test_infer_timestamp_with_tz():
    now_raw = datetime.utcnow()
    now_utc = pytz.utc.localize(now_raw)
    assert now_utc.tzinfo == pytz.UTC
    assert dt.infer(now_utc).timezone == str(pytz.UTC)


def test_infer_timedelta():
    assert dt.infer(timedelta(days=3)) == dt.Interval(unit="D")
    assert dt.infer(timedelta(hours=5)) == dt.Interval(unit="s")
    assert dt.infer(timedelta(minutes=7)) == dt.Interval(unit="s")
    assert dt.infer(timedelta(seconds=9)) == dt.Interval(unit="s")
    assert dt.infer(timedelta(milliseconds=11)) == dt.Interval(unit="us")
    assert dt.infer(timedelta(microseconds=13)) == dt.Interval(unit="us")

    msg = "Unable to infer interval type from zero value"
    with pytest.raises(ValueError, match=msg):
        dt.infer(timedelta(days=0, seconds=0))

    msg = "Unable to infer interval type from mixed units"
    with pytest.raises(ValueError, match=msg):
        dt.infer(timedelta(days=1, hours=2))
    with pytest.raises(ValueError, match=msg):
        dt.infer(timedelta(days=1, microseconds=2))


# str, pd.Timestamp, datetime, np.datetime64, numbers.Real
@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("2019-01-01", datetime(2019, 1, 1)),
        ("2019-01-01 00:00:00", datetime(2019, 1, 1)),
        ("2019-01-01 01:02:03.000004", datetime(2019, 1, 1, 1, 2, 3, 4)),
        (
            "2019-01-01 01:02:03.000004+00:00",
            datetime(2019, 1, 1, 1, 2, 3, 4, tzinfo=pytz.utc),
        ),
        (
            "2019-01-01 01:02:03.000004+01:00",
            datetime(2019, 1, 1, 1, 2, 3, 4, tzinfo=pytz.FixedOffset(60)),
        ),
        (
            "2019-01-01 01:02:03.000004-01:00",
            datetime(2019, 1, 1, 1, 2, 3, 4, tzinfo=pytz.FixedOffset(-60)),
        ),
        (
            "2019-01-01 01:02:03.000004+01",
            datetime(2019, 1, 1, 1, 2, 3, 4, tzinfo=pytz.FixedOffset(60)),
        ),
        (datetime(2019, 1, 1), datetime(2019, 1, 1)),
        (datetime(2019, 1, 1, 1, 2, 3, 4), datetime(2019, 1, 1, 1, 2, 3, 4)),
        (pd.Timestamp("2019-01-01"), datetime(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 00:00:00"), datetime(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 01:02:03.000004"), datetime(2019, 1, 1, 1, 2, 3, 4)),
        (np.datetime64("2019-01-01"), datetime(2019, 1, 1)),
        (np.datetime64("2019-01-01 01:02:03"), datetime(2019, 1, 1, 1, 2, 3)),
    ],
)
def test_normalize_timestamp(value, expected):
    normalized = dt.normalize(dt.timestamp, value)
    assert normalized == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("2019-01-01", date(2019, 1, 1)),
        ("2019-01-01 00:00:00", date(2019, 1, 1)),
        ("2019-01-01 01:02:03.000004", date(2019, 1, 1)),
        (datetime(2019, 1, 1), date(2019, 1, 1)),
        (datetime(2019, 1, 1, 1, 2, 3, 4), date(2019, 1, 1)),
        (pd.Timestamp("2019-01-01"), date(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 00:00:00"), date(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 01:02:03.000004"), date(2019, 1, 1)),
        (np.datetime64("2019-01-01"), date(2019, 1, 1)),
        (np.datetime64("2019-01-01 01:02:03"), date(2019, 1, 1)),
    ],
)
def test_normalize_date(value, expected):
    normalized = dt.normalize(dt.date, value)
    assert normalized == expected


@pytest.mark.parametrize(
    ("dtype", "value", "expected"),
    [
        (dt.Interval("s"), timedelta(seconds=1), 1),
        (dt.Interval("ms"), timedelta(seconds=1), 1000),
        (dt.Interval("us"), timedelta(seconds=1), 1000000),
        (dt.Interval("ns"), timedelta(seconds=1), 1000000000),
    ],
)
def test_normalize_interval(dtype, value, expected):
    normalized = dt.normalize(dtype, value)
    assert normalized == expected


@pytest.mark.parametrize(
    ("value", "expected_dtype"),
    [
        # numpy types
        (np.int8(5), dt.int8),
        (np.int16(-1), dt.int16),
        (np.int32(2), dt.int32),
        (np.int64(-5), dt.int64),
        (np.uint8(5), dt.uint8),
        (np.uint16(50), dt.uint16),
        (np.uint32(500), dt.uint32),
        (np.uint64(5000), dt.uint64),
        (np.float32(5.5), dt.float32),
        (np.float64(5.55), dt.float64),
        (np.bool_(True), dt.boolean),
        (np.bool_(False), dt.boolean),
        # pandas types
        (
            pd.Timestamp("2015-01-01 12:00:00", tz="US/Eastern"),
            dt.Timestamp("US/Eastern"),
        ),
    ],
)
def test_infer_numpy_scalar(value, expected_dtype):
    assert dt.infer(value) == expected_dtype


@pytest.mark.parametrize(
    ("numpy_dtype", "ibis_dtype"),
    [
        (np.bool_, dt.boolean),
        (np.int8, dt.int8),
        (np.int16, dt.int16),
        (np.int32, dt.int32),
        (np.int64, dt.int64),
        (np.uint8, dt.uint8),
        (np.uint16, dt.uint16),
        (np.uint32, dt.uint32),
        (np.uint64, dt.uint64),
        (np.float16, dt.float16),
        (np.float32, dt.float32),
        (np.float64, dt.float64),
        (np.double, dt.double),
        (np.str_, dt.string),
        (np.datetime64, dt.timestamp),
    ],
)
def test_from_numpy_dtype(numpy_dtype, ibis_dtype):
    numpy_dtype = np.dtype(numpy_dtype)
    assert dt.DataType.from_numpy(numpy_dtype) == ibis_dtype
    assert dt.dtype(numpy_dtype) == ibis_dtype


def test_from_numpy_timedelta():
    if vparse(pytest.importorskip("pyarrow").__version__) < vparse("9"):
        pytest.skip("pyarrow < 9 globally mutates the timedelta64 numpy dtype")

    numpy_dtype = np.dtype(np.timedelta64)
    assert dt.DataType.from_numpy(numpy_dtype) == dt.Interval("s")
    assert dt.dtype(numpy_dtype) == dt.Interval("s")


@pytest.mark.parametrize(
    ("numpy_array", "expected_dtypes"),
    [
        # Explicitly-defined dtype
        (np.array([1, 2, 3], dtype="int8"), (dt.Array(dt.int8),)),
        (np.array([1, 2, 3], dtype="int16"), (dt.Array(dt.int16),)),
        (np.array([1, 2, 3], dtype="int32"), (dt.Array(dt.int32),)),
        (np.array([1, 2, 3], dtype="int64"), (dt.Array(dt.int64),)),
        (np.array([1, 2, 3], dtype="uint8"), (dt.Array(dt.uint8),)),
        (np.array([1, 2, 3], dtype="uint16"), (dt.Array(dt.uint16),)),
        (np.array([1, 2, 3], dtype="uint32"), (dt.Array(dt.uint32),)),
        (np.array([1, 2, 3], dtype="uint64"), (dt.Array(dt.uint64),)),
        (np.array([1.0, 2.0, 3.0], dtype="float32"), (dt.Array(dt.float32),)),
        (np.array([1.0, 2.0, 3.0], dtype="float64"), (dt.Array(dt.float64),)),
        (np.array([True, False, True], dtype="bool"), (dt.Array(dt.boolean),)),
        # Implicit dtype
        # Integer array could be inferred to int64 or int32 depending on system
        (np.array([1, 2, 3]), (dt.Array(dt.int64), dt.Array(dt.int32))),
        (np.array([1.0, 2.0, 3.0]), (dt.Array(dt.float64),)),
        (np.array([np.nan, np.nan, np.nan]), (dt.Array(dt.float64),)),
        (np.array([True, False, True]), (dt.Array(dt.boolean),)),
        (np.array(["1", "2", "3"]), (dt.Array(dt.string),)),
        (
            np.array(
                [
                    pd.Timestamp("2015-01-01 12:00:00"),
                    pd.Timestamp("2015-01-02 12:00:00"),
                    pd.Timestamp("2015-01-03 12:00:00"),
                ]
            ),
            (dt.Array(dt.Timestamp()), dt.Array(dt.Timestamp(scale=6))),
        ),
        # Implied from object dtype
        (np.array([1, 2, 3], dtype=object), (dt.Array(dt.int64),)),
        (np.array([1.0, 2.0, 3.0], dtype=object), (dt.Array(dt.float64),)),
        (np.array([True, False, True], dtype=object), (dt.Array(dt.boolean),)),
        (np.array(["1", "2", "3"], dtype=object), (dt.Array(dt.string),)),
        (
            np.array(
                [
                    pd.Timestamp("2015-01-01 12:00:00"),
                    pd.Timestamp("2015-01-02 12:00:00"),
                    pd.Timestamp("2015-01-03 12:00:00"),
                ],
                dtype=object,
            ),
            (dt.Array(dt.Timestamp()), dt.Array(dt.Timestamp(scale=6))),
        ),
    ],
)
def test_infer_numpy_array(numpy_array, expected_dtypes):
    pandas_series = pd.Series(numpy_array)
    assert dt.infer(numpy_array) in expected_dtypes
    assert dt.infer(pandas_series) in expected_dtypes


def test_normalize_json():
    obj = ["foo", {"bar": ("baz", None, 1.0, 2)}]
    expected = json.dumps(obj)

    assert dt.normalize(dt.json, obj) == expected
    assert dt.normalize(dt.json, expected) == expected

    with pytest.raises(TypeError):
        dt.normalize(dt.json, "invalid")


def test_normalize_none_with_non_nullable_type():
    typ = dt.Int64(nullable=False)
    with pytest.raises(TypeError, match="Cannot convert `None` to non-nullable type"):
        dt.normalize(typ, None)


def test_normalize_non_convertible_boolean():
    typ = dt.boolean
    value = np.array([1, 2, 3])
    with pytest.raises(TypeError, match="Unable to normalize .+ to Boolean"):
        dt.normalize(typ, value)


@pytest.mark.parametrize("bits", [8, 16, 32, 64])
@pytest.mark.parametrize("kind", ["uint", "int"])
def test_normalize_non_convertible_int(kind, bits):
    typ = getattr(dt, f"{kind}{bits:d}")
    with pytest.raises(TypeError, match="Unable to normalize .+ to U?Int"):
        dt.normalize(typ, "not convertible")


@pytest.mark.parametrize("typename", ["float32", "float64"])
def test_normalize_non_convertible_float(typename):
    typ = getattr(dt, typename)
    with pytest.raises(TypeError, match="Unable to normalize .+ to Float"):
        dt.normalize(typ, "not convertible")


@pytest.mark.parametrize(
    ("value", "dtype", "expected"),
    [
        (1, dt.Decimal(), "1"),
        (1.0, dt.Decimal(), "1"),
        (1.0, dt.Decimal(2, 1), "1.0"),
        (1.0, dt.Decimal(2, 0), "1"),
        (1.0, dt.Decimal(4, 3), "1.000"),
        (12, dt.Decimal(6, 3), "12.000"),
        (12.1234, dt.Decimal(7, 5), "12.12340"),
        (True, dt.Decimal(4, 0), "1"),
        (True, dt.Decimal(4, 3), "1.000"),
        (False, dt.Decimal(4, 0), "0"),
        (decimal.Decimal("1.1"), dt.Decimal(76, 38), "1.1" + "0" * 37),
    ],
)
def test_normalize_decimal(value, dtype, expected):
    assert str(dt.normalize(dtype, value)) == expected


def test_normalize_decimal_invalid():
    with pytest.raises(TypeError):
        dt.normalize(dt.Decimal(4, 2), "invalid")
    with pytest.raises(TypeError):
        dt.normalize(dt.Decimal(4, 2), 1234)
    with pytest.raises(TypeError):
        dt.normalize(12.1234, dt.Decimal(6, 2))



from __future__ import annotations

import io
from datetime import date
from operator import methodcaller

import numpy as np
import pandas as pd
import pytest
from packaging.version import parse as vparse

import ibis
import ibis.expr.datatypes as dt
from ibis.backends.pandas import Backend
from ibis.backends.pandas.tests.conftest import TestConf as tm
from ibis.common.annotations import ValidationError
from ibis.legacy.udf.vectorized import reduction


@pytest.fixture(scope="session")
def sort_kind():
    return "mergesort"


default = pytest.mark.parametrize("default", [ibis.NA, ibis.literal("a")])
row_offset = pytest.mark.parametrize("row_offset", list(map(ibis.literal, [-1, 1, 0])))
range_offset = pytest.mark.parametrize(
    "range_offset",
    [
        ibis.interval(days=1),
        2 * ibis.interval(days=1),
        -2 * ibis.interval(days=1),
    ],
)


@pytest.fixture
def row_window():
    return ibis.window(following=0, order_by="plain_int64")


@pytest.fixture
def range_window():
    return ibis.window(following=0, order_by="plain_datetimes_naive")


@default
@row_offset
def test_lead(t, df, row_offset, default, row_window):
    con = ibis.pandas.connect()
    expr = t.dup_strings.lead(row_offset, default=default).over(row_window)
    result = expr.execute()
    expected = df.dup_strings.shift(con.execute(-row_offset))
    if default is not ibis.NA:
        expected = expected.fillna(con.execute(default))
    tm.assert_series_equal(result, expected.rename("tmp"))


@default
@row_offset
def test_lag(t, df, row_offset, default, row_window):
    con = ibis.pandas.connect()
    expr = t.dup_strings.lag(row_offset, default=default).over(row_window)
    result = expr.execute()
    expected = df.dup_strings.shift(con.execute(row_offset))
    if default is not ibis.NA:
        expected = expected.fillna(con.execute(default))
    tm.assert_series_equal(result, expected.rename("tmp"))


@default
@range_offset
def test_lead_delta(t, df, range_offset, default, range_window):
    con = ibis.pandas.connect()
    expr = t.dup_strings.lead(range_offset, default=default).over(range_window)
    result = expr.execute()
    expected = (
        df[["plain_datetimes_naive", "dup_strings"]]
        .set_index("plain_datetimes_naive")
        .squeeze()
        .shift(freq=con.execute(-range_offset))
        .reindex(df.plain_datetimes_naive)
        .reset_index(drop=True)
    )
    if default is not ibis.NA:
        expected = expected.fillna(con.execute(default))
    tm.assert_series_equal(result, expected.rename("tmp"))


@default
@range_offset
def test_lag_delta(t, df, range_offset, default, range_window):
    con = ibis.pandas.connect()
    expr = t.dup_strings.lag(range_offset, default=default).over(range_window)
    result = expr.execute()

    expected = (
        df[["plain_datetimes_naive", "dup_strings"]]
        .set_index("plain_datetimes_naive")
        .squeeze()
        .shift(freq=con.execute(range_offset))
        .reindex(df.plain_datetimes_naive)
        .reset_index(drop=True)
    )
    if default is not ibis.NA:
        expected = expected.fillna(con.execute(default))
    tm.assert_series_equal(result, expected.rename("tmp"))


def test_first(t, df):
    expr = t.dup_strings.first()
    result = expr.execute()
    expected = df.dup_strings.iat[0]
    assert result == expected


def test_last(t, df):
    expr = t.dup_strings.last()
    result = expr.execute()
    expected = df.dup_strings.iat[-1]
    assert result == expected


def test_first_and_last_over_window(t):
    def simple_window_ops(table):
        w = ibis.window(
            order_by=[table.plain_uint64, table.dup_ints],
            preceding=1,
            following=0,
        )
        return table.mutate(
            x_first=t.plain_uint64.first().over(w),
            x_last=t.plain_uint64.last().over(w),
            y_first=t.dup_ints.first().over(w),
            y_last=t.dup_ints.last().over(w),
        )

    assert simple_window_ops(t).execute() is not None


def test_group_by_mutate_analytic(t, df):
    gb = t.group_by(t.dup_strings)
    expr = gb.mutate(
        first_value=t.plain_int64.first(),
        last_value=t.plain_strings.last(),
        avg_broadcast=t.plain_float64 - t.plain_float64.mean(),
        delta=(t.plain_int64 - t.plain_int64.lag())
        / (t.plain_float64 - t.plain_float64.lag()),
    )
    result = expr.execute()

    gb = df.groupby("dup_strings")
    expected = df.assign(
        last_value=gb.plain_strings.transform("last"),
        first_value=gb.plain_int64.transform("first"),
        avg_broadcast=df.plain_float64 - gb.plain_float64.transform("mean"),
        delta=(
            (df.plain_int64 - gb.plain_int64.shift(1))
            / (df.plain_float64 - gb.plain_float64.shift(1))
        ),
    )

    tm.assert_frame_equal(result[expected.columns], expected)


def test_players(players, players_df):
    lagged = players.mutate(pct=lambda t: t.G - t.G.lag())
    expected = players_df.assign(
        pct=players_df.G - players_df.groupby("playerID").G.shift(1)
    )
    cols = expected.columns.tolist()
    result = lagged.execute()[cols].sort_values(cols).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


def test_batting_filter_mean(batting, batting_df):
    expr = batting[batting.G > batting.G.mean()]
    result = expr.execute()
    expected = batting_df[batting_df.G > batting_df.G.mean()].reset_index(drop=True)
    tm.assert_frame_equal(result[expected.columns], expected)


def test_batting_zscore(players, players_df):
    expr = players.mutate(g_z=lambda t: (t.G - t.G.mean()) / t.G.std())

    gb = players_df.groupby("playerID")
    expected = players_df.assign(
        g_z=(players_df.G - gb.G.transform("mean")) / gb.G.transform("std")
    )
    cols = expected.columns.tolist()
    result = expr.execute()[cols].sort_values(cols).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


def test_batting_avg_change_in_games_per_year(players, players_df):
    expr = players.mutate(
        delta=lambda t: (t.G - t.G.lag()) / (t.yearID - t.yearID.lag())
    )

    gb = players_df.groupby("playerID")
    expected = players_df.assign(
        delta=(players_df.G - gb.G.shift(1)) / (players_df.yearID - gb.yearID.shift(1))
    )

    cols = expected.columns.tolist()
    result = expr.execute()[cols].sort_values(cols).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


@pytest.mark.xfail(raises=AssertionError, reason="NYI")
def test_batting_most_hits(players, players_df):
    expr = players.mutate(
        hits_rank=lambda t: t.H.rank().over(
            ibis.cumulative_window(order_by=ibis.desc(t.H))
        )
    )
    result = expr.execute()
    hits_rank = players_df.groupby("playerID").H.rank(method="min", ascending=False)
    expected = players_df.assign(hits_rank=hits_rank)
    tm.assert_frame_equal(result[expected.columns], expected)


def test_batting_quantile(players, players_df):
    expr = players.mutate(hits_quantile=lambda t: t.H.quantile(0.25))
    hits_quantile = players_df.groupby("playerID").H.transform("quantile", 0.25)
    expected = players_df.assign(hits_quantile=hits_quantile)
    cols = expected.columns.tolist()
    result = expr.execute()[cols].sort_values(cols).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


def test_batting_approx_median(players, players_df):
    expr = players.mutate(hits_median=lambda t: t.H.approx_median())
    hits_median = players_df.groupby("playerID").H.transform("median")
    expected = players_df.assign(hits_median=hits_median)
    cols = expected.columns.tolist()
    result = expr.execute()[cols].sort_values(cols).reset_index(drop=True)
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize("op", ["sum", "mean", "min", "max"])
def test_batting_specific_cumulative(batting, batting_df, op, sort_kind):
    ibis_method = methodcaller(f"cum{op}", order_by=batting.yearID)
    expr = ibis_method(batting.G)
    result = expr.execute().astype("float64")

    pandas_method = methodcaller(op)
    expected = pandas_method(
        batting_df[["G", "yearID"]].sort_values("yearID", kind=sort_kind).G.expanding()
    ).reset_index(drop=True)
    tm.assert_series_equal(result, expected.rename("tmp"))


def test_batting_cumulative(batting, batting_df, sort_kind):
    expr = batting.mutate(
        more_values=lambda t: t.G.sum().over(ibis.cumulative_window(order_by=t.yearID))
    )
    result = expr.execute()

    columns = ["G", "yearID"]
    more_values = (
        batting_df[columns]
        .sort_values("yearID", kind=sort_kind)
        .G.expanding()
        .sum()
        .astype("int64")
    )
    expected = batting_df.assign(more_values=more_values)
    tm.assert_frame_equal(result[expected.columns], expected)


def test_batting_cumulative_partitioned(batting, batting_df, sort_kind):
    group_by = "playerID"
    order_by = "yearID"

    t = batting
    expr = t.G.sum().over(ibis.cumulative_window(order_by=order_by, group_by=group_by))
    expr = t.mutate(cumulative=expr)
    result = expr.execute()

    columns = [group_by, order_by, "G"]
    expected = (
        batting_df[columns]
        .set_index(order_by)
        .groupby(group_by)
        .G.expanding()
        .sum()
        .rename("cumulative")
    )

    tm.assert_series_equal(
        result.set_index([group_by, order_by]).sort_index().cumulative,
        expected.sort_index().astype("int64"),
    )


def test_batting_rolling(batting, batting_df, sort_kind):
    expr = batting.mutate(
        more_values=lambda t: t.G.sum().over(ibis.trailing_window(5, order_by=t.yearID))
    )
    result = expr.execute()

    columns = ["G", "yearID"]
    more_values = (
        batting_df[columns]
        .sort_values("yearID", kind=sort_kind)
        .G.rolling(6, min_periods=1)
        .sum()
        .astype("int64")
    )
    expected = batting_df.assign(more_values=more_values)
    tm.assert_frame_equal(result[expected.columns], expected)


def test_batting_rolling_partitioned(batting, batting_df, sort_kind):
    t = batting
    group_by = "playerID"
    order_by = "yearID"
    expr = t.G.sum().over(
        ibis.trailing_window(3, order_by=t[order_by], group_by=t[group_by])
    )
    expr = t.mutate(rolled=expr)
    result = expr.execute()

    columns = [group_by, order_by, "G"]
    expected = (
        batting_df[columns]
        .set_index(order_by)
        .groupby(group_by)
        .G.rolling(4, min_periods=1)
        .sum()
        .rename("rolled")
    )

    tm.assert_series_equal(
        result.set_index([group_by, order_by]).sort_index().rolled,
        expected.sort_index().astype("int64"),
    )


def test_scalar_broadcasting(batting, batting_df):
    expr = batting.mutate(demeaned=batting.G - batting.G.mean())
    result = expr.execute()
    expected = batting_df.assign(demeaned=batting_df.G - batting_df.G.mean())
    tm.assert_frame_equal(result, expected)


def test_mutate_with_window_after_join(sort_kind):
    left_df = pd.DataFrame(
        {
            "ints": [0, 1, 2],
            "strings": ["a", "b", "c"],
            "dates": pd.date_range("20170101", periods=3),
        }
    )
    right_df = pd.DataFrame(
        {
            "group": [0, 1, 2] * 3,
            "value": [0, 1, np.nan, 3, 4, np.nan, 6, 7, 8],
        }
    )
    con = Backend().connect({"left": left_df, "right": right_df})
    left, right = map(con.table, ("left", "right"))

    joined = left.outer_join(right, left.ints == right.group)
    proj = joined[left, right.value]
    expr = proj.group_by("ints").mutate(sum=proj.value.sum())
    result = expr.execute()
    expected = pd.DataFrame(
        {
            "dates": pd.concat([left_df.dates] * 3)
            .sort_values(kind=sort_kind)
            .reset_index(drop=True),
            "ints": [0] * 3 + [1] * 3 + [2] * 3,
            "strings": ["a"] * 3 + ["b"] * 3 + ["c"] * 3,
            "value": [0.0, 3.0, 6.0, 1.0, 4.0, 7.0, np.nan, np.nan, 8.0],
            "sum": [9.0] * 3 + [12.0] * 3 + [8.0] * 3,
        }
    )
    tm.assert_frame_equal(result[expected.columns], expected)


def test_mutate_scalar_with_window_after_join():
    left_df = pd.DataFrame({"ints": range(3)})
    right_df = pd.DataFrame(
        {
            "group": [0, 1, 2] * 3,
            "value": [0, 1, np.nan, 3, 4, np.nan, 6, 7, 8],
        }
    )
    con = Backend().connect({"left": left_df, "right": right_df})
    left, right = map(con.table, ("left", "right"))

    joined = left.outer_join(right, left.ints == right.group)
    proj = joined[left, right.value]
    expr = proj.mutate(sum=proj.value.sum(), const=1)
    result = expr.execute()
    expected = pd.DataFrame(
        {
            "ints": [0] * 3 + [1] * 3 + [2] * 3,
            "value": [0.0, 3.0, 6.0, 1.0, 4.0, 7.0, np.nan, np.nan, 8.0],
            "sum": [29.0] * 9,
            "const": np.ones(9, dtype="int8"),
        }
    )
    tm.assert_frame_equal(result[expected.columns], expected)


def test_project_scalar_after_join():
    left_df = pd.DataFrame({"ints": range(3)})
    right_df = pd.DataFrame(
        {
            "group": [0, 1, 2] * 3,
            "value": [0, 1, np.nan, 3, 4, np.nan, 6, 7, 8],
        }
    )
    con = ibis.pandas.connect({"left": left_df, "right": right_df})
    left, right = map(con.table, ("left", "right"))

    joined = left.outer_join(right, left.ints == right.group)
    proj = joined[left, right.value]
    expr = proj[proj.value.sum().name("sum"), ibis.literal(1).name("const")]
    result = expr.execute()
    expected = pd.DataFrame(
        {
            "sum": [29.0] * 9,
            "const": np.ones(9, dtype="int8"),
        }
    )
    tm.assert_frame_equal(result[expected.columns], expected)


def test_project_list_scalar():
    df = pd.DataFrame({"ints": range(3)})
    con = ibis.pandas.connect({"df": df})
    table = con.table("df")
    expr = table.mutate(res=table.ints.quantile([0.5, 0.95]))
    result = expr.execute()

    expected = pd.Series([[1.0, 1.9] for _ in range(3)], name="res")
    tm.assert_series_equal(result.res, expected)


@pytest.mark.parametrize(
    "index",
    [
        pytest.param(lambda time: None, id="no_index"),
        pytest.param(lambda time: time, id="index"),
    ],
)
def test_window_with_preceding_expr(index):
    time = pd.date_range("20180101", "20180110")
    start = 2
    data = np.arange(start, start + len(time))
    df = pd.DataFrame({"value": data, "time": time}, index=index(time))
    client = ibis.pandas.connect({"df": df})
    t = client.table("df")
    expected = (
        df.set_index("time")
        .value.rolling("3d", closed="both")
        .mean()
        .reset_index(drop=True)
    )
    expected.index.name = None
    day = ibis.interval(days=1)
    window = ibis.trailing_window(3 * day, order_by=t.time)
    expr = t.value.mean().over(window)
    result = expr.execute()
    tm.assert_series_equal(result, expected.rename("mean"))


@pytest.mark.xfail(
    condition=vparse("1.4") <= vparse(pd.__version__) < vparse("1.4.2"),
    raises=ValueError,
    reason="https://github.com/pandas-dev/pandas/pull/44068",
)
def test_window_with_mlb():
    index = pd.date_range("20170501", "20170507")
    data = np.random.randn(len(index), 3)
    df = (
        pd.DataFrame(data, columns=list("abc"), index=index)
        .rename_axis("time")
        .reset_index(drop=False)
    )
    client = ibis.pandas.connect({"df": df})
    t = client.table("df")
    rows_with_mlb = ibis.rows_with_max_lookback(5, ibis.interval(days=10))
    expr = t.mutate(
        sum=lambda df: df.a.sum().over(
            ibis.trailing_window(rows_with_mlb, order_by="time", group_by="b")
        )
    )
    result = expr.execute()
    expected = df.set_index("time")
    gb_df = (
        expected.groupby(["b"])["a"]
        .rolling("10d", closed="both")
        .apply(lambda s: s.iloc[-5:].sum(), raw=False)
        .sort_index(level=["time"])
        .reset_index(drop=True)
    )
    expected = expected.reset_index(drop=False).assign(sum=gb_df)
    tm.assert_frame_equal(result, expected)

    rows_with_mlb = ibis.rows_with_max_lookback(5, 10)
    with pytest.raises(ValidationError):
        t.mutate(
            sum=lambda df: df.a.sum().over(
                ibis.trailing_window(rows_with_mlb, order_by="time")
            )
        )


def test_window_grouping_key_has_scope(t, df):
    param = ibis.param(dt.string)
    window = ibis.window(group_by=t.dup_strings + param)
    expr = t.plain_int64.mean().over(window)
    result = expr.execute(params={param: "a"})
    expected = df.groupby(df.dup_strings + "a").plain_int64.transform("mean")
    tm.assert_series_equal(result, expected.rename("mean"))


def test_window_on_and_by_key_as_window_input(t, df):
    order_by = "plain_int64"
    group_by = "dup_ints"
    control = "plain_float64"

    row_window = ibis.trailing_window(order_by=order_by, group_by=group_by, preceding=1)

    # Test built-in function

    tm.assert_series_equal(
        t[order_by].count().over(row_window).execute(),
        t[control].count().over(row_window).execute(),
        check_names=False,
    )

    tm.assert_series_equal(
        t[group_by].count().over(row_window).execute(),
        t[control].count().over(row_window).execute(),
        check_names=False,
    )

    # Test UDF

    @reduction(input_type=[dt.int64], output_type=dt.int64)
    def count(v):
        return len(v)

    @reduction(input_type=[dt.int64, dt.int64], output_type=dt.int64)
    def count_both(v1, v2):
        return len(v1)

    tm.assert_series_equal(
        count(t[order_by]).over(row_window).execute(),
        t[control].count().over(row_window).execute(),
        check_names=False,
    )

    tm.assert_series_equal(
        count(t[group_by]).over(row_window).execute(),
        t[control].count().over(row_window).execute(),
        check_names=False,
    )

    tm.assert_series_equal(
        count_both(t[group_by], t[order_by]).over(row_window).execute(),
        t[control].count().over(row_window).execute(),
        check_names=False,
    )


@pytest.mark.parametrize(
    "group_by,order_by",
    [
        (None, None),
        # Enable this after #2395 is merged
        # (None, 'plain_datetimes_utc'),
        ("dup_ints", None),
        ("dup_ints", "plain_datetimes_utc"),
    ],
)
def test_rolling_window_udf_nan_and_non_numeric(t, group_by, order_by):
    # Test that rolling window can be performed on
    # (1) A column that contains NaN values
    # (2) A non-numeric column
    # (3) A non-numeric column that contains NaN value

    t = t.mutate(nan_int64=t["plain_int64"])
    t = t.mutate(nan_int64=None)

    @reduction(input_type=[dt.int64], output_type=dt.int64)
    def count_int64(v):
        return len(v)

    @reduction(input_type=[dt.timestamp], output_type=dt.int64)
    def count_timestamp(v):
        return len(v)

    @reduction(input_type=[t["map_of_strings_integers"].type()], output_type=dt.int64)
    def count_complex(v):
        return len(v)

    window = ibis.trailing_window(preceding=1, order_by=order_by, group_by=group_by)

    result_nan = count_int64(t["nan_int64"]).over(window).execute()
    result_non_numeric = (
        count_timestamp(t["plain_datetimes_utc"]).over(window).execute()
    )
    result_nan_non_numeric = (
        count_timestamp(t["map_of_strings_integers"]).over(window).execute()
    )
    expected = t["plain_int64"].count().over(window).execute()

    tm.assert_series_equal(result_nan, expected, check_names=False)
    tm.assert_series_equal(result_non_numeric, expected, check_names=False)
    tm.assert_series_equal(result_nan_non_numeric, expected, check_names=False)


@pytest.fixture
def events():
    df = pd.DataFrame(
        {
            "event_id": [1] * 4 + [2] * 6 + [3] * 2,
            "measured_on": map(
                pd.Timestamp,
                map(
                    date,
                    [2021] * 12,
                    [6] * 4 + [5] * 6 + [7] * 2,
                    range(1, 13),
                ),
            ),
            "measurement": np.nan,
        }
    )
    df.at[1, "measurement"] = 5.0
    df.at[4, "measurement"] = 42.0
    df.at[5, "measurement"] = 42.0
    df.at[7, "measurement"] = 11.0
    return df


def test_bfill(events):
    con = ibis.pandas.connect({"t": events})
    t = con.table("t")

    win = ibis.window(
        group_by=t.event_id, order_by=ibis.desc(t.measured_on), following=0
    )
    grouped = t.mutate(grouper=t.measurement.count().over(win))

    expr = (
        grouped.group_by([grouped.event_id, grouped.grouper])
        .mutate(bfill=grouped.measurement.max())
        .order_by("measured_on")
    )
    result = expr.execute().reset_index(drop=True)

    expected_raw = """\
event_id measured_on  measurement  grouper  bfill
       2  2021-05-05         42.0        3   42.0
       2  2021-05-06         42.0        2   42.0
       2  2021-05-07          NaN        1   11.0
       2  2021-05-08         11.0        1   11.0
       2  2021-05-09          NaN        0    NaN
       2  2021-05-10          NaN        0    NaN
       1  2021-06-01          NaN        1    5.0
       1  2021-06-02          5.0        1    5.0
       1  2021-06-03          NaN        0    NaN
       1  2021-06-04          NaN        0    NaN
       3  2021-07-11          NaN        0    NaN
       3  2021-07-12          NaN        0    NaN"""
    expected = pd.read_csv(
        io.StringIO(expected_raw),
        sep=r"\s+",
        header=0,
        parse_dates=["measured_on"],
    )
    tm.assert_frame_equal(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING

import sqlglot as sg

import ibis
import ibis.common.exceptions as com
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis.backends.impala import ddl
from ibis.backends.impala.ddl import AlterTable, InsertSelect

if TYPE_CHECKING:
    import pandas as pd


class ImpalaTable(ir.Table):
    """A physical table in the Impala-Hive metastore."""

    @property
    def _qualified_name(self) -> str:
        op = self.op()
        return sg.table(op.name, catalog=op.namespace.database).sql(dialect="hive")

    @property
    def _unqualified_name(self) -> str:
        return self.op().name

    @property
    def _client(self):
        return self.op().source

    @property
    def _database(self) -> str:
        return self.op().namespace.database

    def compute_stats(self, incremental=False):
        """Invoke Impala COMPUTE STATS command on the table."""
        return self._client.compute_stats(
            self.op().name, database=self._database, incremental=incremental
        )

    def invalidate_metadata(self):
        self._client.invalidate_metadata(self.op().name, database=self._database)

    def refresh(self):
        self._client.refresh(self.op().name, database=self._database)

    def metadata(self):
        """Return results of `DESCRIBE FORMATTED` statement."""
        return self._client.describe_formatted(self.op().name, database=self._database)

    describe_formatted = metadata

    def files(self):
        """Return results of SHOW FILES statement."""
        return self._client.show_files(self.op().name, database=self._database)

    def drop(self):
        """Drop the table from the database."""
        self._client.drop_table_or_view(self.op().name, database=self._database)

    def truncate(self):
        self._client.truncate_table(self.op().name, database=self._database)

    def insert(
        self,
        obj=None,
        overwrite=False,
        partition=None,
        values=None,
        validate=True,
    ):
        """Insert into an Impala table.

        Parameters
        ----------
        obj
            Table expression or DataFrame
        overwrite
            If True, will replace existing contents of table
        partition
            For partitioned tables, indicate the partition that's being
            inserted into, either with an ordered list of partition keys or a
            dict of partition field name to value. For example for the
            partition (year=2007, month=7), this can be either (2007, 7) or
            {'year': 2007, 'month': 7}.
        values
            Unsupported and unused
        validate
            If True, do more rigorous validation that schema of table being
            inserted is compatible with the existing table

        Examples
        --------
        Append to an existing table

        >>> t.insert(table_expr)  # quartodoc: +SKIP # doctest: +SKIP

        Completely overwrite contents

        >>> t.insert(table_expr, overwrite=True)  # quartodoc: +SKIP # doctest: +SKIP

        """
        if values is not None:
            raise NotImplementedError

        if not isinstance(obj, ir.Table):
            obj = ibis.memtable(obj)

        self._client._run_pre_execute_hooks(obj)

        expr = obj
        if validate:
            existing_schema = self.schema()
            insert_schema = expr.schema()
            if not insert_schema.equals(existing_schema):
                _validate_compatible(insert_schema, existing_schema)

        if partition is not None:
            partition_schema = self.partition_schema()
            partition_schema_names = frozenset(partition_schema.names)
            expr = expr.select(
                [
                    column
                    for column in expr.columns
                    if column not in partition_schema_names
                ]
            )
        else:
            partition_schema = None

        statement = InsertSelect(
            self._qualified_name,
            self._client.compile(expr),
            partition=partition,
            partition_schema=partition_schema,
            overwrite=overwrite,
        )
        self._client._safe_exec_sql(statement.compile())
        return self

    def load_data(self, path, overwrite=False, partition=None):
        """Load data into an Impala table.

        Parameters
        ----------
        path
            Data to load
        overwrite
            Overwrite the existing data in the entire table or indicated
            partition
        partition
            If specified, the partition must already exist

        """
        if partition is not None:
            partition_schema = self.partition_schema()
        else:
            partition_schema = None

        stmt = ddl.LoadData(
            self._qualified_name,
            path,
            partition=partition,
            partition_schema=partition_schema,
            overwrite=overwrite,
        )

        self._client._safe_exec_sql(stmt.compile())
        return self

    @property
    def name(self) -> str:
        return self.op().name

    @property
    def is_partitioned(self):
        """True if the table is partitioned."""
        return self.metadata().is_partitioned

    def partition_schema(self):
        """Return the schema for the partition columns."""
        schema = self.schema()
        result = self.partitions()

        partition_fields = []
        for col in result.columns:
            if col not in schema:
                break
            partition_fields.append((col, schema[col]))

        return sch.Schema(dict(partition_fields))

    def add_partition(self, spec, location=None):
        """Add a new table partition.

        Partition parameters can be set in a single DDL statement or you can
        use `alter_partition` to set them after the fact.
        """
        part_schema = self.partition_schema()
        stmt = ddl.AddPartition(
            self._qualified_name, spec, part_schema, location=location
        )
        self._client._safe_exec_sql(stmt)
        return self

    def alter(
        self,
        location=None,
        format=None,
        tbl_properties=None,
        serde_properties=None,
    ):
        """Change settings and parameters of the table.

        Parameters
        ----------
        location
            For partitioned tables, you may want the alter_partition function
        format
            Table format
        tbl_properties
            Table properties
        serde_properties
            Serialization/deserialization properties

        """

        def _run_ddl(**kwds):
            stmt = AlterTable(self._qualified_name, **kwds)
            self._client._safe_exec_sql(stmt)
            return self

        return self._alter_table_helper(
            _run_ddl,
            location=location,
            format=format,
            tbl_properties=tbl_properties,
            serde_properties=serde_properties,
        )

    def set_external(self, is_external=True):
        """Toggle the `EXTERNAL` table property."""
        self.alter(tbl_properties={"EXTERNAL": is_external})

    def alter_partition(
        self,
        spec,
        location=None,
        format=None,
        tbl_properties=None,
        serde_properties=None,
    ):
        """Change settings and parameters of an existing partition.

        Parameters
        ----------
        spec
            The partition keys for the partition being modified
        location
            Location of the partition
        format
            Table format
        tbl_properties
            Table properties
        serde_properties
            Serialization/deserialization properties

        """
        part_schema = self.partition_schema()

        def _run_ddl(**kwds):
            stmt = ddl.AlterPartition(self._qualified_name, spec, part_schema, **kwds)
            self._client._safe_exec_sql(stmt)
            return self

        return self._alter_table_helper(
            _run_ddl,
            location=location,
            format=format,
            tbl_properties=tbl_properties,
            serde_properties=serde_properties,
        )

    def _alter_table_helper(self, f, **alterations):
        results = []
        for k, v in alterations.items():
            if v is None:
                continue
            result = f(**{k: v})
            results.append(result)
        return results

    def drop_partition(self, spec):
        """Drop an existing table partition."""
        part_schema = self.partition_schema()
        stmt = ddl.DropPartition(self._qualified_name, spec, part_schema)
        self._client._safe_exec_sql(stmt)
        return self

    def partitions(self):
        """Return information about the table's partitions.

        Raises an exception if the table is not partitioned.
        """
        return self._client.list_partitions(self._qualified_name)

    def stats(self) -> pd.DataFrame:
        """Return results of `SHOW TABLE STATS`.

        If not partitioned, contains only one row.

        Returns
        -------
        DataFrame
            Table statistics

        """
        return self._client.table_stats(self._qualified_name)

    def column_stats(self) -> pd.DataFrame:
        """Return results of `SHOW COLUMN STATS`.

        Returns
        -------
        DataFrame
            Column statistics

        """
        return self._client.column_stats(self._qualified_name)


# ----------------------------------------------------------------------
# ORM-ish usability layer


class ScalarFunction:
    def drop(self):
        pass


class AggregateFunction:
    def drop(self):
        pass


def _validate_compatible(from_schema, to_schema):
    if set(from_schema.names) != set(to_schema.names):
        raise com.IbisInputError("Schemas have different names")

    for name in from_schema:
        lt = from_schema[name]
        rt = to_schema[name]
        if not lt.castable(rt):
            raise com.IbisInputError(f"Cannot safely cast {lt!r} to {rt!r}")



from __future__ import annotations

from typing import Optional

import pytest

import ibis
import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.common.annotations import ValidationError
from ibis.common.patterns import EqualTo

t = ibis.table([("a", "int64")], name="t")

true = ir.literal(True)
false = ir.literal(False)
two = ir.literal(2)
three = ir.literal(3)


class Expr:
    def __init__(self, op):
        self.op = op


class Base(ops.Node):
    def to_expr(self):
        return Expr(self)


class Name(Base):
    name: str


class NamedValue(Base):
    value: int
    name: Name


class Values(Base):
    lst: tuple[ops.Node, ...]


one = NamedValue(value=1, name=Name("one"))
two = NamedValue(value=2, name=Name("two"))
three = NamedValue(value=3, name=Name("three"))
values = Values((one, two, three))


def test_node_base():
    assert hasattr(one, "__slots__")
    assert not hasattr(one, "__dict__")
    assert one.__args__ == (1, Name("one"))
    assert values.__args__ == ((one, two, three),)

    calls = []
    returns = {
        Name("one"): "Name_one",
        Name("two"): "Name_two",
        Name("three"): "Name_three",
        NamedValue(1, Name("one")): "NamedValue_1_one",
        NamedValue(2, Name("two")): "NamedValue_2_two",
        NamedValue(3, Name("three")): "NamedValue_3_three",
        values: "final",
    }

    def record(node, _, *args, **kwargs):
        calls.append((node, args, kwargs))
        return returns[node]

    results = values.map(record)

    assert results == returns
    assert calls == [
        (Name("one"), (), {"name": "one"}),
        (Name("two"), (), {"name": "two"}),
        (Name("three"), (), {"name": "three"}),
        (one, (), {"value": 1, "name": "Name_one"}),
        (two, (), {"value": 2, "name": "Name_two"}),
        (three, (), {"value": 3, "name": "Name_three"}),
        (
            values,
            (),
            {"lst": ("NamedValue_1_one", "NamedValue_2_two", "NamedValue_3_three")},
        ),
    ]


def test_node_substitution():
    class Aliased(Base):
        arg: ops.Node
        name: str

    ketto = Aliased(one, "ketto")

    first_rule = EqualTo(Name("one")) >> Name("zero")
    second_rule = EqualTo(two) >> ketto

    new_values = values.replace(first_rule | second_rule)
    expected = Values((NamedValue(value=1, name=Name("zero")), ketto, three))

    assert expected == new_values


def test_value_annotations():
    class Op1(ops.Value):
        arg: ops.Value

        dtype = dt.int64
        shape = ds.scalar

    class Op2(ops.Value):
        arg: ops.Value[dt.Any, ds.Any]

        dtype = dt.int64
        shape = ds.scalar

    class Op3(ops.Value):
        arg: ops.Value[dt.Integer, ds.Any]

        dtype = dt.int64
        shape = ds.scalar

    class Op4(ops.Value):
        arg: ops.Value[dt.Integer, ds.Scalar]

        dtype = dt.int64
        shape = ds.scalar

    assert Op1(1).arg.dtype == dt.int8
    assert Op2(1).arg.dtype == dt.int8
    assert Op3(1).arg.dtype == dt.int8
    assert Op4(1).arg.dtype == dt.int8


def test_operation_definition():
    class Logarithm(ir.Expr):
        pass

    class Log(ops.Node):
        arg: ops.Value[dt.Float64, ds.Any]
        base: Optional[ops.Value[dt.Float64, ds.Any]] = None

        def to_expr(self):
            return Logarithm(self)

    assert Log(1, base=2).arg == ops.Literal(1, dtype=dt.float64)
    assert Log(1, base=2).base == ops.Literal(2, dtype=dt.float64)
    assert Log(arg=10).arg == ops.Literal(10, dtype=dt.float64)
    assert Log(arg=10).base is None

    assert isinstance(Log(arg=100).to_expr(), Logarithm)


def test_instance_of_operation():
    class MyOperation(ops.Node):
        arg: ir.IntegerValue

        def to_expr(self):
            return ir.IntegerScalar(self)

    MyOperation(ir.literal(5))

    with pytest.raises(ValidationError):
        MyOperation(ir.literal("string"))


def test_array_input():
    class MyOp(ops.Value):
        value: ops.Value[dt.Array[dt.Float64], ds.Any]
        dtype = rlz.dtype_like("value")
        shape = rlz.shape_like("value")

    raw_value = [1.0, 2.0, 3.0]
    op = MyOp(raw_value)

    expected = ibis.literal(raw_value)
    assert op.value == expected.op()


def test_custom_table_expr():
    class MyTable(ir.Table):
        pass

    class SpecialTable(ops.UnboundTable):
        def to_expr(self):
            return MyTable(self)

    node = SpecialTable(name="foo", schema=ibis.schema([("a", "int64")]))
    expr = node.to_expr()
    assert isinstance(expr, MyTable)


def test_too_many_or_too_few_args_not_allowed():
    class DummyOp(ops.Value):
        arg: ops.Value

    with pytest.raises(ValidationError):
        DummyOp(1, 2)

    with pytest.raises(ValidationError):
        DummyOp()


def test_getitem_on_column_is_error():
    t = ibis.table(dict(a="int"))

    with pytest.raises(TypeError, match="#ibis-for-pandas-users"):
        t.a[0]

    with pytest.raises(TypeError, match="#ibis-for-pandas-users"):
        t.a[:1]


def test_operation_class_aliases():
    assert ops.ValueOp is ops.Value
    assert ops.UnaryOp is ops.Unary
    assert ops.BinaryOp is ops.Binary
    assert ops.WindowOp is ops.Window
    assert ops.AnalyticOp is ops.Analytic


def test_expression_class_aliases():
    assert ir.Table is ir.Table
    assert ir.ValueExpr is ir.Value
    assert ir.ScalarExpr is ir.Scalar
    assert ir.ColumnExpr is ir.Column
    assert ir.AnyValue is ir.Value
    assert ir.AnyScalar is ir.Scalar
    assert ir.AnyColumn is ir.Column



from __future__ import annotations

import pytest

from ibis.backends.base import _get_backend_names

all_backends = list(_get_backend_names())


@pytest.mark.notimpl(all_backends)
def test_notimpl(con):
    raise Exception


@pytest.mark.notimpl(all_backends, raises=None)
def test_notimpl_raises_none(con):
    raise Exception


@pytest.mark.notimpl(all_backends, raises=(None, None))
def test_notimpl_raises_none_tuple(con):
    raise Exception


@pytest.mark.notimpl(all_backends, raises=(Exception, None))
def test_notimpl_raises_tuple_exception_none(con):
    raise Exception


@pytest.mark.notyet(all_backends)
def test_notyet(con):
    raise Exception


@pytest.mark.notyet(all_backends, raises=None)
def test_notyet_raises_none(con):
    raise Exception


@pytest.mark.notyet(all_backends, raises=(None, None))
def test_notyet_raises_none_tuple(con):
    raise Exception


@pytest.mark.notyet(all_backends, raises=(Exception, None))
def test_notyet_raises_tuple_exception_none(con):
    raise Exception


@pytest.mark.never(all_backends, reason="because I said so")
def test_never(con):
    raise Exception


@pytest.mark.never(all_backends, raises=None, reason="because I said so")
def test_never_raises_none(con):
    raise Exception


@pytest.mark.never(all_backends, raises=(None, None), reason="because I said so")
def test_never_raises_none_tuple(con):
    raise Exception


@pytest.mark.never(all_backends, raises=(Exception, None), reason="because I said so")
def test_never_raises_tuple_exception_none(con):
    raise Exception
